# -*- coding: utf-8 -*-
"""This module acts as a bridge to the sector models from the controller

The :class:`SectorModel` exposes several key methods for running wrapped
sector models.  To add a sector model to an instance of the framework,
first implement :class:`SectorModel`.

Key Functions
=============
This class performs several key functions which ease the integration of sector
models into the system-of-systems framework.

The user must implement the various abstract functions throughout the class to
provide an interface to the sector model, which can be called upon by the
framework. From the model's perspective, :class:`SectorModel` provides a bridge
from the sector-specific problem representation to the general representation
which allows reasoning across infrastructure systems.

The key functions include

* converting input/outputs to/from geographies/temporal resolutions
* converting control vectors from the decision layer of the framework, to
  asset Interventions specific to the sector model
* returning scaler/vector values to the framework to enable measurements of
  performance, particularly for the purposes of optimisation and rule-based
  approaches

"""
import logging
import os
from abc import ABC, abstractmethod

import importlib
from smif.parameters import ModelParameters

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class SectorModel(ABC):
    """A representation of the sector model with inputs and outputs

    """
    def __init__(self):
        self._model_name = None

        self.interventions = []
        self.system = []

        self._inputs = ModelParameters({})
        self._outputs = ModelParameters({})

        self.logger = logging.getLogger(__name__)

    def validate(self):
        """Validate that this SectorModel has been set up with sufficient data
        to run
        """
        pass

    @property
    def name(self):
        """The name of the sector model

        Returns
        =======
        str
            The name of the sector model
        """
        return self._model_name

    @name.setter
    def name(self, value):
        self._model_name = value

    @property
    def inputs(self):
        """The inputs to the model

        The inputs should be specified in a list.  For example::

                - name: eletricity_price
                  spatial_resolution: GB
                  temporal_resolution: annual

        Arguments
        =========
        value : list
            A list of dicts of inputs to the model.
            These includes parameters, assets and exogenous data

        Returns
        =======
        :class:`smif.parameters.ModelInputs`

        """
        return self._inputs

    @inputs.setter
    def inputs(self, value=None):
        if value is not None:
            assert isinstance(value, list)
        else:
            value = []

        self._inputs = ModelParameters(value)

    @property
    def outputs(self):
        """The outputs from the model

        Arguments
        =========
        value : list
            A list of dicts of outputs from the model.
            This may include results
            and metrics

        Returns
        =======
        :class:`smif.parameters.ModelParameters`

        """
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        if value is not None:
            assert isinstance(value, list)
        else:
            value = []

        self._outputs = ModelParameters(value)

    @property
    def intervention_names(self):
        """The names of the interventions

        Returns
        =======
        list
            A list of the names of the interventions
        """
        return [intervention['name'] for intervention in self.interventions]

    @abstractmethod
    def initialise(self, initial_conditions):
        """Implement this method to set up the model system

        Arguments
        ---------
        initial_conditions: list
            A list of past Interventions, with build dates and locations as
            necessary to specify the infrastructure system to be modelled.
        """
        pass

    @abstractmethod
    def simulate(self, decisions, state, data):
        """Implement this method to run the model

        Arguments
        ---------
        decisions: list
            A list of :py:class:Intervention to apply to the modelled system
        state: list
            A list of :py:class:StateData to update the state of the modelled system
        data: dict
            A dictionary of the format:
            ``data[parameter] = [SpaceTimeValue(region, interval, value, units), ...]``
        Returns
        -------
        dict
            A dictionary of the format:
            ``results[parameter] = [SpaceTimeValue(region, interval, value, units), ...]``

        Notes
        -----
        In the results returned from the :py:meth:`simulate` method:

        ``interval``
            should reference an id from the interval set corresponding to
            the output parameter, as specified in model configuration
        ``region``
            should reference a region name from the region set corresponding to
            the output parameter, as specified in model configuration

        """
        pass

    @abstractmethod
    def extract_obj(self, results):
        """Implement this method to return a scalar value objective function

        This method should take the results from the output of the
        :py:meth:`simulate` method, process the results,
        and return a scalar value which can be used as a component of
        the objective function by the decision layer

        Arguments
        =========
        results : dict
            A nested dict of the results from the :py:meth:`simulate` method

        Returns
        =======
        float
            A scalar component generated from the simulation model results
        """
        pass


class SectorModelBuilder(object):
    """Build the components that make up a sectormodel from the configuration

    Parameters
    ----------
    name : str
        The name of the sector model

    """

    def __init__(self, name):
        self._sector_model_name = name
        self._sector_model = None
        self.logger = logging.getLogger(__name__)

    def load_model(self, model_path, classname):
        """Dynamically load model module

        """
        if os.path.exists(model_path):
            self.logger.info("Importing run module from %s", model_path)

            spec = importlib.util.spec_from_file_location(
                self._sector_model_name, model_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            klass = module.__dict__[classname]

            self._sector_model = klass()
            self._sector_model.name = self._sector_model_name

        else:
            msg = "Cannot find '{}' for the '{}' model".format(
                model_path, self._sector_model_name)
            raise FileNotFoundError(msg)

    def create_initial_system(self, initial_conditions):
        """Set up model with initial system
        """
        msg = "Sector model must be loaded before creating initial system"
        assert self._sector_model is not None, msg

        self._sector_model.initialise(initial_conditions)

    def add_inputs(self, input_dict):
        """Add inputs to the sector model
        """
        msg = "Sector model must be loaded before adding inputs"
        assert self._sector_model is not None, msg

        self._sector_model.inputs = input_dict

    def add_outputs(self, output_dict):
        """Add outputs to the sector model
        """
        msg = "Sector model must be loaded before adding outputs"
        assert self._sector_model is not None, msg

        self._sector_model.outputs = output_dict

    def add_interventions(self, intervention_list):
        """Add interventions to the sector model

        Parameters
        ----------
        intervention_list : list
            A list of dicts of interventions

        """
        msg = "Sector model must be loaded before adding interventions"
        assert self._sector_model is not None, msg

        self._sector_model.interventions = intervention_list

    def validate(self):
        """Check and/or assert that the sector model is correctly set up
        - should raise errors if invalid
        """
        assert self._sector_model is not None, "Sector model not loaded"
        self._sector_model.validate()
        return True

    def finish(self):
        """Validate and return the sector model
        """
        self.validate()
        return self._sector_model



"""In this module, we implement the conversion across space and time

The :class:`SpaceTimeConvertor` is instantiated with data to convert,
and the names of the four source and destination spatio-temporal resolutions.

The :meth:`~SpaceTimeConvertor.convert` method returns a new list of
:class:`smif.SpaceTimeValue` namedtuples for passing to a sector model.
"""
from collections import OrderedDict
import logging
from smif.convert.interval import TimeSeries
from smif import SpaceTimeValue

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class SpaceTimeConvertor(object):
    """Handles the conversion of time and space for a list of values

    Arguments
    ---------
    data: list
        A list of :class:`smif.SpaceTimeValue`
    from_spatial: str
        The name of the spatial resolution of the data
    to_spatial: str
        The name of the required spatial resolution
    from_temporal: str
        The name of the temporal resolution of the data
    to_temporal: str
        The name of the required temproal resolution
    region_register: :class:`smif.convert.area.RegionRegister`
        A fully populated register of the models' regions
    interval_register: :class:`smif.convert.interval.TimeIntervalRegister`
        A fully populated register of the models' intervals

    Notes
    -----
    Future development requires using a data object which allows multiple views
    upon the values across the three dimensions of time, space and units. This
    will then allow more efficient conversion across any one of these dimensions
    while holding the others constant.  One option could be
    :class:`collections.ChainMap`.

    """

    def __init__(self, data,
                 from_spatial, to_spatial,
                 from_temporal, to_temporal,
                 region_register, interval_register):
        self.logger = logging.getLogger(__name__)

        self._check_uniform_units(data)

        self.data = data
        self.from_spatial = from_spatial
        self.to_spatial = to_spatial
        self.from_temporal = from_temporal
        self.to_temporal = to_temporal

        self.regions = region_register
        self.intervals = interval_register

    @property
    def data_regions(self):
        data_by_regions = self._regionalise_data(self.data)
        return set(data_by_regions.keys())

    @property
    def data_by_region(self):
        return self._regionalise_data(self.data)

    def _check_uniform_units(self, data):
        units = []
        for entry in data:
            units.append(entry.units)
        if len(set(units)) > 1:
            msg = "SpaceTimeConvertor cannot handle multiple units for conversion"
            raise NotImplementedError(msg)

    @staticmethod
    def _regionalise_data(data):
        """

        Parameters
        ----------
        data: list
            A list of :class:`smif.SpaceTimeValue`

        Returns
        -------
        :class:`collections.OrderedDict`
            A dictionary where the key is a region, and the value is a list of
            :class:`smif.SpaceTimeValue`

        """
        data_by_region = OrderedDict()
        for entry in data:
            if entry.region not in data_by_region:
                data_by_region[entry.region] = [entry]
            else:
                data_by_region[entry.region].append(entry)
        return data_by_region

    @staticmethod
    def _intervalise_data(data):
        """

        Parameters
        ----------
        data: list
            A list of :class:`smif.SpaceTimeValue`

        Returns
        -------
        :class:`collections.OrderedDict`
            A dictionary where the key is an interval, and the value is a list of
            :class:`smif.SpaceTimeValue`

        """
        data_by_intervals = OrderedDict()
        for entry in data:
            if entry.interval not in data_by_intervals:
                data_by_intervals[entry.interval] = [entry]
            else:
                data_by_intervals[entry.interval].append(entry)
        return data_by_intervals

    def convert(self):
        """Convert the data according to the parameters passed
        to the SpaceTimeConvertor

        Returns
        -------
        list
            A list of :class:`smif.SpaceTimeValue`
        """
        assert self.from_spatial in self.regions.region_set_names
        assert self.to_spatial in self.regions.region_set_names
        assert self.from_temporal in self.intervals.interval_set_names
        assert self.to_temporal in self.intervals.interval_set_names

        if self._convert_intervals_required() and self._convert_regions_required():
            self.logger.debug("Converting intervals and regions")
            interval_data = self._loop_over_regions(self.data)
            data = self._loop_over_intervals(interval_data)
        elif self._convert_intervals_required():
            self.logger.debug("Converting intervals only")
            data = self._loop_over_regions(self.data)
        elif self._convert_regions_required():
            self.logger.debug("Converting regions only")
            data = self._loop_over_intervals(self.data)
        else:
            self.logger.debug("No conversion required, passing data through")
            data = self.data

        return data

    def _loop_over_regions(self, data=None):
        """
        """
        converted_data = []
        data_by_region = self._regionalise_data(data)
        num_regions = len(set(data_by_region.keys()))
        if num_regions > 1:
            for region, region_data in data_by_region.items():
                converted_data.extend(self._convert_time(region_data))
        elif num_regions == 1:
            converted_data = self._convert_time(data)
        return converted_data

    def _loop_over_intervals(self, data=None):
        """
        """
        converted_data = []
        data_by_intervals = self._intervalise_data(data)
        num_intervals = len(set(data_by_intervals.keys()))
        if num_intervals > 1:
            for interval, interval_data in data_by_intervals.items():
                converted_data.extend(self._convert_space(interval_data))
        elif num_intervals == 1:
            converted_data = self._convert_space(data)
        return converted_data

    def _convert_regions_required(self):
        """Returns True if it is necessary to convert over space

        Returns
        -------
        bool
        """
        if self.from_spatial == self.to_spatial:
            return False
        elif self.from_spatial != self.to_spatial:
            return True
        else:
            msg = "Cannot determine if region conversion is required"
            raise ValueError(msg)

    def _convert_intervals_required(self):
        """Returns True if it is necessary to convert over time

        Returns
        -------
        bool
        """
        if self.from_temporal == self.to_temporal:
            return False
        elif self.from_temporal != self.to_temporal:
            return True
        else:
            msg = "Cannot determine if time conversion is required"
            raise ValueError(msg)

    def _convert_space(self, data):
        """Wraps the call to the area conversion

        Parameters
        ----------
        data: list
            A list of :class:`smif.SpaceTimeValue` for one interval

        Returns
        -------
        data: list
            A list of :class:`smif.SpaceTimeValue` for one interval
        """
        timeseries_data = {}

        interval = data[0].interval
        units = data[0].units
        msg = "Converting regions for interval %s with units %s"
        self.logger.debug(msg, interval, units)

        for entry in data:
            timeseries_data[entry.region] = entry.value

        converted_data = self.regions.convert(timeseries_data,
                                              self.from_spatial,
                                              self.to_spatial)
        return [SpaceTimeValue(region_name, interval, converted_value, units)
                for region_name, converted_value in converted_data.items()]

    def _convert_time(self, data):
        """Wraps the call to the interval conversion

        Parameters
        ----------
        data: list
            A list of :class:`smif.SpaceTimeValue` for one region

        Returns
        -------
        data: list
            A list of :class:`smif.SpaceTimeValue` for one region
        """

        timeseries_data = []

        region = data[0].region
        units = data[0].units

        for entry in data:
            timeseries_data.append({'id': entry.interval,
                                    'value': entry.value})

        timeseries = TimeSeries(timeseries_data)
        converted_data = self.intervals.convert(timeseries,
                                                self.from_temporal,
                                                self.to_temporal)

        return [SpaceTimeValue(region, entry['id'], entry['value'], units)
                for entry in converted_data]



from pytest import approx, fixture, raises
from smif import SpaceTimeValue
from smif.convert import SpaceTimeConvertor
from smif.convert.area import RegionRegister
from smif.convert.interval import TimeIntervalRegister
from test_area import regions_half_squares, regions_rect
from test_interval import (months, one_day, remap_months, seasons,
                           twenty_four_hours)


@fixture(scope='function')
def data_remap():
    data = [SpaceTimeValue('a', '1', 30+31+31, 'days'),
            SpaceTimeValue('a', '2', 28+31+30, 'days'),
            SpaceTimeValue('a', '3', 31+31+30, 'days'),
            SpaceTimeValue('a', '4', 30+31+31, 'days')]
    return data


@fixture(scope='function')
def expected_stv_remap():
    data = [SpaceTimeValue('a', '1_0', 30.666666666, 'days'),
            SpaceTimeValue('a', '1_1', 29.666666666, 'days'),
            SpaceTimeValue('a', '1_2', 29.666666666, 'days'),
            SpaceTimeValue('a', '1_3', 29.666666666, 'days'),
            SpaceTimeValue('a', '1_4', 30.666666666, 'days'),
            SpaceTimeValue('a', '1_5', 30.666666666, 'days'),
            SpaceTimeValue('a', '1_6', 30.666666666, 'days'),
            SpaceTimeValue('a', '1_7', 30.666666666, 'days'),
            SpaceTimeValue('a', '1_8', 30.666666666, 'days'),
            SpaceTimeValue('a', '1_9', 30.666666666, 'days'),
            SpaceTimeValue('a', '1_10', 30.666666666, 'days'),
            SpaceTimeValue('a', '1_11', 30.666666666, 'days')]
    return data


@fixture(scope='function')
def monthly_data():
    """[31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    """
    data = [SpaceTimeValue('a', '1_0', 31, 'days'),
            SpaceTimeValue('a', '1_1', 28, 'days'),
            SpaceTimeValue('a', '1_2', 31, 'days'),
            SpaceTimeValue('a', '1_3', 30, 'days'),
            SpaceTimeValue('a', '1_4', 31, 'days'),
            SpaceTimeValue('a', '1_5', 30, 'days'),
            SpaceTimeValue('a', '1_6', 31, 'days'),
            SpaceTimeValue('a', '1_7', 31, 'days'),
            SpaceTimeValue('a', '1_8', 30, 'days'),
            SpaceTimeValue('a', '1_9', 31, 'days'),
            SpaceTimeValue('a', '1_10', 30, 'days'),
            SpaceTimeValue('a', '1_11', 31, 'days')]
    return data


class TestSpaceTimeConvertor_Utils:

    def test_instantiation(self):
        SpaceTimeConvertor([], None, None, None, None, None, None)

    def test_conversion_not_required(self):
        convertor = SpaceTimeConvertor([], 'a', 'a', 'x', 'x', None, None)
        assert convertor._convert_intervals_required() is False
        assert convertor._convert_regions_required() is False

    def test_conversion_is_required(self):
        convertor = SpaceTimeConvertor([], 'a', 'b', 'x', 'y', None, None)
        assert convertor._convert_intervals_required() is True
        assert convertor._convert_regions_required() is True

    def test_multiple_units_raises_notimplemented(self):
        data = [SpaceTimeValue('a', '1_0', 31, 'days'),
                SpaceTimeValue('a', '1_2', 31, 'minutes'),
                SpaceTimeValue('b', '1_1', 28, 'seconds')]
        with raises(NotImplementedError):
            SpaceTimeConvertor(data, 'a', 'a', 'x', 'x', None, None)

    def test_data_by_regions(self):
        data = [SpaceTimeValue('a', '1_0', 31, 'days'),
                SpaceTimeValue('a', '1_2', 31, 'days'),
                SpaceTimeValue('b', '1_1', 28, 'days')]
        convertor = SpaceTimeConvertor(data, 'a', 'a', 'x', 'x', None, None)

        expected = {'a': [SpaceTimeValue('a', '1_0', 31, 'days'),
                          SpaceTimeValue('a', '1_2', 31, 'days')],
                    'b': [SpaceTimeValue('b', '1_1', 28, 'days')]}

        assert convertor.data_by_region == expected


class TestSpaceTimeConvertor_TimeOnly:

    def test_one_region_pass_through_time(self, months, seasons, regions_half_squares):
        """Only one region, 12 months, neither space nor time conversion is required

        """

        data = [SpaceTimeValue('a', '1_0', 31, 'days'),
                SpaceTimeValue('a', '1_1', 28, 'days'),
                SpaceTimeValue('a', '1_2', 31, 'days'),
                SpaceTimeValue('a', '1_3', 30, 'days'),
                SpaceTimeValue('a', '1_4', 31, 'days'),
                SpaceTimeValue('a', '1_5', 30, 'days'),
                SpaceTimeValue('a', '1_6', 31, 'days'),
                SpaceTimeValue('a', '1_7', 31, 'days'),
                SpaceTimeValue('a', '1_8', 30, 'days'),
                SpaceTimeValue('a', '1_9', 31, 'days'),
                SpaceTimeValue('a', '1_10', 30, 'days'),
                SpaceTimeValue('a', '1_11', 31, 'days')]

        intervals = TimeIntervalRegister()
        intervals.register(months, 'months')

        regions = RegionRegister()
        regions.register(regions_half_squares)

        convertor = SpaceTimeConvertor(data,
                                       'half_squares',
                                       'half_squares',
                                       'months',
                                       'months',
                                       regions,
                                       intervals)

        assert convertor.data_regions == set(['a'])
        assert convertor.data_by_region == {'a': data}

        actual = convertor.convert()

        expected = [SpaceTimeValue('a', '1_0', 31, 'days'),
                    SpaceTimeValue('a', '1_1', 28, 'days'),
                    SpaceTimeValue('a', '1_2', 31, 'days'),
                    SpaceTimeValue('a', '1_3', 30, 'days'),
                    SpaceTimeValue('a', '1_4', 31, 'days'),
                    SpaceTimeValue('a', '1_5', 30, 'days'),
                    SpaceTimeValue('a', '1_6', 31, 'days'),
                    SpaceTimeValue('a', '1_7', 31, 'days'),
                    SpaceTimeValue('a', '1_8', 30, 'days'),
                    SpaceTimeValue('a', '1_9', 31, 'days'),
                    SpaceTimeValue('a', '1_10', 30, 'days'),
                    SpaceTimeValue('a', '1_11', 31, 'days')]

        assert actual == expected

    def test_one_region_time_aggregation(self, months, seasons, regions_half_squares):
        """Only one region, time aggregation is required
        """

        data = [SpaceTimeValue('a', '1_0', 31, 'days'),
                SpaceTimeValue('a', '1_1', 28, 'days'),
                SpaceTimeValue('a', '1_2', 31, 'days'),
                SpaceTimeValue('a', '1_3', 30, 'days'),
                SpaceTimeValue('a', '1_4', 31, 'days'),
                SpaceTimeValue('a', '1_5', 30, 'days'),
                SpaceTimeValue('a', '1_6', 31, 'days'),
                SpaceTimeValue('a', '1_7', 31, 'days'),
                SpaceTimeValue('a', '1_8', 30, 'days'),
                SpaceTimeValue('a', '1_9', 31, 'days'),
                SpaceTimeValue('a', '1_10', 30, 'days'),
                SpaceTimeValue('a', '1_11', 31, 'days')]

        intervals = TimeIntervalRegister()
        intervals.register(months, 'months')
        intervals.register(seasons, 'seasons')

        regions = RegionRegister()
        regions.register(regions_half_squares)

        convertor = SpaceTimeConvertor(data,
                                       'half_squares',
                                       'half_squares',
                                       'months',
                                       'seasons',
                                       regions,
                                       intervals)
        assert convertor.data_regions == set(['a'])
        assert convertor.data_by_region == {'a': data}

        actual = convertor.convert()

        expected = [SpaceTimeValue('a', 'winter', 31. + 31 + 28, 'days'),
                    SpaceTimeValue('a', 'spring', 31. + 30 + 31, 'days'),
                    SpaceTimeValue('a', 'summer', 30. + 31 + 31, 'days'),
                    SpaceTimeValue('a', 'autumn', 30. + 31 + 30, 'days')]
        assert isinstance(actual, list)
        for entry in actual:
            assert isinstance(entry, SpaceTimeValue)

        for act, exp in zip(actual, expected):
            assert act.region == exp.region
            assert act.interval == exp.interval
            assert act.value == approx(exp.value)
            assert act.units == exp.units

    def test_two_region_time_aggregation(self, months, seasons, regions_half_squares):
        """Two regions, time aggregation by region is required
        """

        data = [SpaceTimeValue('a', '1_0', 31, 'days'),
                SpaceTimeValue('a', '1_1', 28, 'days'),
                SpaceTimeValue('a', '1_2', 31, 'days'),
                SpaceTimeValue('a', '1_3', 30, 'days'),
                SpaceTimeValue('a', '1_4', 31, 'days'),
                SpaceTimeValue('a', '1_5', 30, 'days'),
                SpaceTimeValue('a', '1_6', 31, 'days'),
                SpaceTimeValue('a', '1_7', 31, 'days'),
                SpaceTimeValue('a', '1_8', 30, 'days'),
                SpaceTimeValue('a', '1_9', 31, 'days'),
                SpaceTimeValue('a', '1_10', 30, 'days'),
                SpaceTimeValue('a', '1_11', 31, 'days'),
                SpaceTimeValue('b', '1_0', 31+1, 'days'),
                SpaceTimeValue('b', '1_1', 28+1, 'days'),
                SpaceTimeValue('b', '1_2', 31+1, 'days'),
                SpaceTimeValue('b', '1_3', 30+1, 'days'),
                SpaceTimeValue('b', '1_4', 31+1, 'days'),
                SpaceTimeValue('b', '1_5', 30+1, 'days'),
                SpaceTimeValue('b', '1_6', 31+1, 'days'),
                SpaceTimeValue('b', '1_7', 31+1, 'days'),
                SpaceTimeValue('b', '1_8', 30+1, 'days'),
                SpaceTimeValue('b', '1_9', 31+1, 'days'),
                SpaceTimeValue('b', '1_10', 30+1, 'days'),
                SpaceTimeValue('b', '1_11', 31+1, 'days')]

        expected_a = [SpaceTimeValue('a', '1_0', 31, 'days'),
                      SpaceTimeValue('a', '1_1', 28, 'days'),
                      SpaceTimeValue('a', '1_2', 31, 'days'),
                      SpaceTimeValue('a', '1_3', 30, 'days'),
                      SpaceTimeValue('a', '1_4', 31, 'days'),
                      SpaceTimeValue('a', '1_5', 30, 'days'),
                      SpaceTimeValue('a', '1_6', 31, 'days'),
                      SpaceTimeValue('a', '1_7', 31, 'days'),
                      SpaceTimeValue('a', '1_8', 30, 'days'),
                      SpaceTimeValue('a', '1_9', 31, 'days'),
                      SpaceTimeValue('a', '1_10', 30, 'days'),
                      SpaceTimeValue('a', '1_11', 31, 'days')]

        expected_b = [SpaceTimeValue('b', '1_0', 31+1, 'days'),
                      SpaceTimeValue('b', '1_1', 28+1, 'days'),
                      SpaceTimeValue('b', '1_2', 31+1, 'days'),
                      SpaceTimeValue('b', '1_3', 30+1, 'days'),
                      SpaceTimeValue('b', '1_4', 31+1, 'days'),
                      SpaceTimeValue('b', '1_5', 30+1, 'days'),
                      SpaceTimeValue('b', '1_6', 31+1, 'days'),
                      SpaceTimeValue('b', '1_7', 31+1, 'days'),
                      SpaceTimeValue('b', '1_8', 30+1, 'days'),
                      SpaceTimeValue('b', '1_9', 31+1, 'days'),
                      SpaceTimeValue('b', '1_10', 30+1, 'days'),
                      SpaceTimeValue('b', '1_11', 31+1, 'days')]

        intervals = TimeIntervalRegister()
        intervals.register(months, 'months')
        intervals.register(seasons, 'seasons')

        regions = RegionRegister()
        regions.register(regions_half_squares)

        convertor = SpaceTimeConvertor(data,
                                       'half_squares',
                                       'half_squares',
                                       'months',
                                       'seasons',
                                       regions,
                                       intervals)
        assert convertor.data_regions == set(['a', 'b'])
        assert convertor.data_by_region['a'] == expected_a
        assert convertor.data_by_region['b'] == expected_b

        actual = convertor.convert()

        expected = [SpaceTimeValue('a', 'winter', 31. + 31 + 28, 'days'),
                    SpaceTimeValue('a', 'spring', 31. + 30 + 31, 'days'),
                    SpaceTimeValue('a', 'summer', 30. + 31 + 31, 'days'),
                    SpaceTimeValue('a', 'autumn', 30. + 31 + 30, 'days'),
                    SpaceTimeValue('b', 'winter', 31. + 31 + 28 + 3, 'days'),
                    SpaceTimeValue('b', 'spring', 31. + 30 + 31 + 3, 'days'),
                    SpaceTimeValue('b', 'summer', 30. + 31 + 31 + 3, 'days'),
                    SpaceTimeValue('b', 'autumn', 30. + 31 + 30 + 3, 'days')]

        for act, exp in zip(actual, expected):
            assert act.region == exp.region
            assert act.interval == exp.interval
            assert act.value == approx(exp.value)
            assert act.units == exp.units

    def test_one_region_convert_from_hour_to_day(self, regions_half_squares,
                                                 twenty_four_hours, one_day):
        """One region, time aggregation required
        """

        data = [SpaceTimeValue('a', '1_0', 1, 'days'),
                SpaceTimeValue('a', '1_1', 1, 'days'),
                SpaceTimeValue('a', '1_2', 1, 'days'),
                SpaceTimeValue('a', '1_3', 1, 'days'),
                SpaceTimeValue('a', '1_4', 1, 'days'),
                SpaceTimeValue('a', '1_5', 1, 'days'),
                SpaceTimeValue('a', '1_6', 1, 'days'),
                SpaceTimeValue('a', '1_7', 1, 'days'),
                SpaceTimeValue('a', '1_8', 1, 'days'),
                SpaceTimeValue('a', '1_9', 1, 'days'),
                SpaceTimeValue('a', '1_10', 1, 'days'),
                SpaceTimeValue('a', '1_11', 1, 'days'),
                SpaceTimeValue('a', '1_12', 1, 'days'),
                SpaceTimeValue('a', '1_13', 1, 'days'),
                SpaceTimeValue('a', '1_14', 1, 'days'),
                SpaceTimeValue('a', '1_15', 1, 'days'),
                SpaceTimeValue('a', '1_16', 1, 'days'),
                SpaceTimeValue('a', '1_17', 1, 'days'),
                SpaceTimeValue('a', '1_18', 1, 'days'),
                SpaceTimeValue('a', '1_19', 1, 'days'),
                SpaceTimeValue('a', '1_20', 1, 'days'),
                SpaceTimeValue('a', '1_21', 1, 'days'),
                SpaceTimeValue('a', '1_22', 1, 'days'),
                SpaceTimeValue('a', '1_23', 1, 'days')]

        regions = RegionRegister()
        regions.register(regions_half_squares)

        intervals = TimeIntervalRegister()
        intervals.register(twenty_four_hours, 'hourly_day')
        intervals.register(one_day, 'one_day')

        convertor = SpaceTimeConvertor(data,
                                       'half_squares',
                                       'half_squares',
                                       'hourly_day',
                                       'one_day',
                                       regions,
                                       intervals)

        assert convertor.data_regions == set(['a'])
        assert convertor.data_by_region['a'] == data

        actual = convertor.convert()

        expected = [SpaceTimeValue('a', 'one_day', 24, 'days')]

        assert actual == expected

    def test_remap_timeslices_to_months(self,
                                        months,
                                        expected_stv_remap,
                                        remap_months,
                                        data_remap,
                                        regions_half_squares):
        """One region, time remapping required
        """
        timeslice_data = data_remap

        intervals = TimeIntervalRegister()
        intervals.register(months, 'months')
        intervals.register(remap_months, 'remap_months')

        regions = RegionRegister()
        regions.register(regions_half_squares)

        convertor = SpaceTimeConvertor(timeslice_data,
                                       'half_squares',
                                       'half_squares',
                                       'remap_months',
                                       'months',
                                       regions,
                                       intervals)

        assert convertor.data_regions == set(['a'])
        assert convertor.data_by_region['a'] == timeslice_data

        actual = convertor.convert()
        expected = expected_stv_remap

        assert len(actual) == len(expected)

        for act, exp in zip(actual, expected):
            assert act.region == exp.region
            assert act.interval == exp.interval
            assert act.value == approx(exp.value)
            assert act.units == exp.units

    def test_remap_months_to_timeslices(self,
                                        months,
                                        monthly_data,
                                        remap_months,
                                        data_remap,
                                        regions_half_squares):
        """One region, time remapping required
        """
        timeslice_data = monthly_data

        intervals = TimeIntervalRegister()
        intervals.register(months, 'months')
        intervals.register(remap_months, 'remap_months')

        regions = RegionRegister()
        regions.register(regions_half_squares)

        convertor = SpaceTimeConvertor(timeslice_data,
                                       'half_squares',
                                       'half_squares',
                                       'months',
                                       'remap_months',
                                       regions,
                                       intervals)

        actual = convertor.convert()
        expected = data_remap

        assert len(actual) == len(expected)

        for act, exp in zip(actual, expected):
            assert act.region == exp.region
            assert act.interval == exp.interval
            assert act.value == approx(exp.value)
            assert act.units == exp.units


class TestSpaceTimeConvertor_RegionOnly:

    def test_half_to_one_region_pass_through_time(self, months,
                                                  regions_half_squares,
                                                  regions_rect):
        """Convert from half a region to one region, pass through time

        """

        data = [SpaceTimeValue('a', '1_0', 0.5, 'days'),
                SpaceTimeValue('a', '1_1', 0.5, 'days'),
                SpaceTimeValue('a', '1_2', 0.5, 'days'),
                SpaceTimeValue('a', '1_3', 0.5, 'days'),
                SpaceTimeValue('a', '1_4', 0.5, 'days'),
                SpaceTimeValue('a', '1_5', 0.5, 'days'),
                SpaceTimeValue('a', '1_6', 0.5, 'days'),
                SpaceTimeValue('a', '1_7', 0.5, 'days'),
                SpaceTimeValue('a', '1_8', 0.5, 'days'),
                SpaceTimeValue('a', '1_9', 0.5, 'days'),
                SpaceTimeValue('a', '1_10', 0.5, 'days'),
                SpaceTimeValue('a', '1_11', 0.5, 'days'),
                SpaceTimeValue('b', '1_0', 0.5, 'days'),
                SpaceTimeValue('b', '1_1', 0.5, 'days'),
                SpaceTimeValue('b', '1_2', 0.5, 'days'),
                SpaceTimeValue('b', '1_3', 0.5, 'days'),
                SpaceTimeValue('b', '1_4', 0.5, 'days'),
                SpaceTimeValue('b', '1_5', 0.5, 'days'),
                SpaceTimeValue('b', '1_6', 0.5, 'days'),
                SpaceTimeValue('b', '1_7', 0.5, 'days'),
                SpaceTimeValue('b', '1_8', 0.5, 'days'),
                SpaceTimeValue('b', '1_9', 0.5, 'days'),
                SpaceTimeValue('b', '1_10', 0.5, 'days'),
                SpaceTimeValue('b', '1_11', 0.5, 'days')]

        intervals = TimeIntervalRegister()
        intervals.register(months, 'months')

        regions = RegionRegister()
        regions.register(regions_half_squares)
        regions.register(regions_rect)

        convertor = SpaceTimeConvertor(data,
                                       'half_squares',
                                       'rect',
                                       'months',
                                       'months',
                                       regions,
                                       intervals)

        actual = convertor.convert()

        expected = [SpaceTimeValue('zero', '1_0', 1, 'days'),
                    SpaceTimeValue('zero', '1_1', 1, 'days'),
                    SpaceTimeValue('zero', '1_2', 1, 'days'),
                    SpaceTimeValue('zero', '1_3', 1, 'days'),
                    SpaceTimeValue('zero', '1_4', 1, 'days'),
                    SpaceTimeValue('zero', '1_5', 1, 'days'),
                    SpaceTimeValue('zero', '1_6', 1, 'days'),
                    SpaceTimeValue('zero', '1_7', 1, 'days'),
                    SpaceTimeValue('zero', '1_8', 1, 'days'),
                    SpaceTimeValue('zero', '1_9', 1, 'days'),
                    SpaceTimeValue('zero', '1_10', 1, 'days'),
                    SpaceTimeValue('zero', '1_11', 1, 'days')]

        for act, exp in zip(actual, expected):
            print("Actual: {}\nExpected: {}".format(actual, expected))
            assert act.region == exp.region
            assert act.interval == exp.interval
            assert act.value == approx(exp.value)
            assert act.units == exp.units

    def test_one_region_convert_from_hour_to_day(self, regions_half_squares,
                                                 regions_rect,
                                                 one_day):
        """Two regions aggregated to one, one interval
        """

        data = [SpaceTimeValue('a', 'one_day', 24, 'days'),
                SpaceTimeValue('b', 'one_day', 24, 'days')]

        intervals = TimeIntervalRegister()
        intervals.register(one_day, 'one_day')

        regions = RegionRegister()
        regions.register(regions_half_squares)
        regions.register(regions_rect)

        convertor = SpaceTimeConvertor(data,
                                       'half_squares',
                                       'rect',
                                       'one_day',
                                       'one_day',
                                       regions,
                                       intervals)

        actual = convertor.convert()

        expected = [SpaceTimeValue('zero', 'one_day', 48, 'days')]

        assert actual == expected


class TestSpaceTimeConvertorBoth:

    def test_convert_both_space_and_time(self, seasons, months,
                                         regions_half_squares,
                                         regions_rect):

        data = [SpaceTimeValue('a', '1_0', 0.5, 'days'),
                SpaceTimeValue('a', '1_1', 0.5, 'days'),
                SpaceTimeValue('a', '1_2', 0.5, 'days'),
                SpaceTimeValue('a', '1_3', 0.5, 'days'),
                SpaceTimeValue('a', '1_4', 0.5, 'days'),
                SpaceTimeValue('a', '1_5', 0.5, 'days'),
                SpaceTimeValue('a', '1_6', 0.5, 'days'),
                SpaceTimeValue('a', '1_7', 0.5, 'days'),
                SpaceTimeValue('a', '1_8', 0.5, 'days'),
                SpaceTimeValue('a', '1_9', 0.5, 'days'),
                SpaceTimeValue('a', '1_10', 0.5, 'days'),
                SpaceTimeValue('a', '1_11', 0.5, 'days'),
                SpaceTimeValue('b', '1_0', 0.5, 'days'),
                SpaceTimeValue('b', '1_1', 0.5, 'days'),
                SpaceTimeValue('b', '1_2', 0.5, 'days'),
                SpaceTimeValue('b', '1_3', 0.5, 'days'),
                SpaceTimeValue('b', '1_4', 0.5, 'days'),
                SpaceTimeValue('b', '1_5', 0.5, 'days'),
                SpaceTimeValue('b', '1_6', 0.5, 'days'),
                SpaceTimeValue('b', '1_7', 0.5, 'days'),
                SpaceTimeValue('b', '1_8', 0.5, 'days'),
                SpaceTimeValue('b', '1_9', 0.5, 'days'),
                SpaceTimeValue('b', '1_10', 0.5, 'days'),
                SpaceTimeValue('b', '1_11', 0.5, 'days')]

        intervals = TimeIntervalRegister()
        intervals.register(months, 'months')
        intervals.register(seasons, 'seasons')

        regions = RegionRegister()
        regions.register(regions_half_squares)
        regions.register(regions_rect)

        convertor = SpaceTimeConvertor(data,
                                       'half_squares',
                                       'rect',
                                       'months',
                                       'seasons',
                                       regions,
                                       intervals)

        actual = convertor.convert()

        expected = [SpaceTimeValue('zero', 'winter', 3, 'days'),
                    SpaceTimeValue('zero', 'spring', 3, 'days'),
                    SpaceTimeValue('zero', 'summer', 3, 'days'),
                    SpaceTimeValue('zero', 'autumn', 3, 'days')]

        for act, exp in zip(actual, expected):
            print("Actual: {}\nExpected: {}".format(actual, expected))
            assert act.region == exp.region
            assert act.interval == exp.interval
            assert act.value == approx(exp.value)
            assert act.units == exp.units



"""This module handles the collection of interventions and assets in a sector.
The set of interventions describes the targets of possible
physical (or non-physical) decisions which can be made in the sector.

An Asset is the general term for an existing component of an infrastructure system.

The difference between an Intervention and an Asset, is that the latter exists
(it has been "built"), whereas the former describes the potential to build an Asset.

The set of assets defines the 'state' of the infrastructure system.

Notes
-----

This module implements:

- initialisation of the set of assets from model config (either a collection of yaml
  text files, or a database)

  - hold generic list of key/values
  - creation of new assets by decision logic (rule-based/optimisation solver)
  - maintain or derive set of possible assets
  - makes the distinction between known-ahead values and build-time values.
    Location and date are specified at build time, while cost and capacity
    are a function of time and location.

- serialisation for passing to models

  - ease of access to full generic data structure

- output list of assets for reporting

  - write out with legible or traceable keys and units for verification and
    understanding

*Terminology*

name:
    A category of infrastructure intervention (e.g. power station, policy)
    which holds default attribute/value pairs. These names can be
    inherited by asset/intervention definitions to reduce the degree of
    duplicate data entry.
asset:
    An instance of an intervention, which represents a single investment
    decisions which will take place, or has taken place.
    Historical interventions are defined as initial conditions, while
    future interventions are listed as pre-specified planning.
    Both historical and future interventions can make use of names to
    ease data entry.  Assets must have ``location``, ``build_date``
    and ``name`` attributes defined.
intervention:
    A potential asset or investment.
    Interventions are defined in the same way as for assets,
    cannot have a ``build_date`` defined.

"""
import hashlib
import json

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class InterventionContainer(object):
    """An container for asset types, interventions and assets.

    An asset's data is set up to be a flexible, plain data structure.

    Parameters
    ----------
    name : str, default=""
        The type of asset, which should be unique across all sectors
    data : dict, default=None
        The dictionary of asset attributes
    sector : str, default=""
        The sector associated with the asset

    """
    def __init__(self, name="", data=None, sector=""):

        assert isinstance(name, str)

        if data is None:
            data = {}

        if name == "" and "name" in data:
            # allow data to set name if none given
            name = data["name"]
        else:
            # otherwise rely on name arg
            data["name"] = name

        self.name = name
        self.data = data

        if sector == "" and "sector" in data:
            # sector is required, may be None
            sector = data["sector"]
        else:
            data["sector"] = sector

        self.sector = sector

        (required, omitted) = self.get_attributes()
        self._validate(required, omitted)

    def get_attributes(self):
        """Override to return two lists, one containing required attributes,
        the other containing omitted attributes

        Returns
        -------
        tuple
            Tuple of lists, one contained required attributes, the other which
            must be omitted
        """
        raise NotImplementedError

    def _validate(self, required_attributes, omitted_attributes):
        """Ensures location is present and no build date is specified

        """
        keys = self.data.keys()
        for expected in required_attributes:
            if expected not in keys:
                msg = "Validation failed due to missing attribute: '{}' in {}"
                raise ValueError(msg.format(expected, str(self)))

        for omitted in omitted_attributes:
            if omitted in keys:
                msg = "Validation failed due to extra attribute: '{}' in {}"
                raise ValueError(msg.format(omitted, str(self)))

    def sha1sum(self):
        """Compute the SHA1 hash of this asset's data

        Returns
        -------
        str
        """
        str_to_hash = str(self).encode('utf-8')
        return hashlib.sha1(str_to_hash).hexdigest()

    def __repr__(self):
        data_str = Asset.deterministic_dict_to_str(self.data)
        return "Asset(\"{}\", {})".format(self.name, data_str)

    def __str__(self):
        return Asset.deterministic_dict_to_str(self.data)

    @staticmethod
    def deterministic_dict_to_str(data):
        """Return a reproducible string representation of any dict

        Parameters
        ----------
        data : dict
            An intervention attributes dictionary

        Returns
        -------
        str
            A reproducible string representation
        """
        return json.dumps(data, sort_keys=True)

    @property
    def sector(self):
        """The name of the sector model this asset is used in.
        """
        return self.data["sector"]

    @sector.setter
    def sector(self, value):
        self.data["sector"] = value


class Intervention(InterventionContainer):
    """An potential investment to send to the logic-layer

    An Intervention, is an investment which has a name (or name),
    other attributes (such as capital cost and economic lifetime),
    and location, but no build date.

    The set of interventions are defined within each sector, and these are
    collected into an :class:`InterventionRegister` when
    a :class:`smif.controller.SosModel` is instantiated by the controller at
    runtime.

    Parameters
    ==========
    name : str, default=""
        The type of asset, which should be unique across all sectors
    data : dict, default=None
        The dictionary of asset attributes
    sector : str, default=""
        The sector associated with the asset

    """
    def get_attributes(self):
        """Ensures location is present and no build date is specified

        """
        return (['name', 'location'], ['build_date'])

    @property
    def location(self):
        """The location of this asset instance (if specified - asset types
        may not have explicit locations)
        """
        return self.data["location"]

    @location.setter
    def location(self, value):
        self.data["location"] = value


class Asset(Intervention):
    """An instance of an intervention with a build date.

    Used to represent pre-specified planning and existing infrastructure assets
    and interventions

    Parameters
    ----------
    name : str, default=""
        The type of asset, which should be unique across all sectors
    data : dict, default=None
        The dictionary of asset attributes
    sector : str, default=""
        The sector associated with the asset

    """
    def get_attributes(self):
        """Ensures location is present and no build date is specified

        """
        return (['name', 'location', 'build_date'], [])

    @property
    def build_date(self):
        """The build date of this asset instance (if specified - asset types
        will not have build dates)
        """
        if "build_date" not in self.data:
            return None
        return self.data["build_date"]

    @build_date.setter
    def build_date(self, value):
        self.data["build_date"] = value


class Register(object):
    """Holds interventions, pre-spec'd planning instructions & existing assets

    - register each asset type/intervention name
    - translate a set of assets representing an initial system into numeric
      representation

    """
    def __init__(self):
        self.assets = {}
        self._names = []
        self._attribute_keys = []
        self._attribute_possible_values = []

    def register(self, asset):
        """Adds a new asset to the collection
        """
        name = asset.data['name']
        self.assets[name] = asset
        self._names.append(asset.data['name'])

        for key in asset.data.keys():
            self._attribute_keys.append(key)

    def __iter__(self):
        for asset in self.assets:
            yield asset


class AssetRegister(Register):
    """Register each asset type

    """

    def register(self, asset):
        if isinstance(asset, Asset):
            pass
        else:
            msg = "You can only register Assets with this register"
            raise TypeError(msg)
        super().register(asset)

    def __len__(self):
        return len(self._names)


class InterventionRegister(Register):
    """The collection of Intervention objects

    An InterventionRegister contains an immutable collection of sector specific
    assets and decision points which can be decided on by the Logic Layer

    * Reads in a collection of interventions defined in each sector model

    * Builds an ordered and immutable collection of interventions

    * Provides interfaces to

      * optimisation/rule-based planning logic

      * SectorModel class model wrappers

    Key functions:

    - outputs a complete list of asset build possibilities (asset type at
      location) which are (potentially) constrained by the pre-specified
      planning instructions and existing infrastructure.

    - translate a binary vector of build instructions
      (e.g. from optimisation routine) into Asset objects with human-readable
      key-value pairs

    - translates an immutable collection of Asset objects into a binary vector
      to pass to the logic-layer.

    Notes
    =====

    *Internal data structures*

    `Intervention_types`
        is a 2D array of integers: each entry is an array
        representing an Intervention type, each integer indexes
        attribute_possible_values

    `attribute_keys`
        is a 1D array of strings

    `attribute_possible_values`
        is a 2D array of simple values, possibly
        (boolean, integer, float, string, tuple). Each entry is a list of
        possible values for the attribute at that index.

    *Invariants*

    - there must be one name and one list of possible values per attribute

    - each Intervention type must list one value for each attribute, and that
      value must be a valid index into the possible_values array

    - each possible_values array should be all of a single type

    """
    def __init__(self):
        super().__init__()
        self._names = {}
        self._numeric_keys = []

    def get_intervention(self, name):
        """Returns the named asset data
        """
        if name in self._names.keys():
            numeric_key = self._names[name]
            return self.numeric_to_intervention(numeric_key)
        else:
            msg = "Intervention '{}' not found in register"
            raise ValueError(msg.format(name))

    def _check_new_intervention(self, intervention):
        """Checks that the asset doesn't exist in the register

        """
        hash_list = []
        for existing_asset in self._numeric_keys:
            hash_list.append(self.numeric_to_intervention(existing_asset).sha1sum())
        if intervention.sha1sum() in hash_list:
            return False
        else:
            return True

    def register(self, intervention):
        """Add a new intervention to the register

        Parameters
        ----------
        intervention : :class:`Intervention`

        """
        if isinstance(intervention, Intervention):
            pass
        else:
            msg = "You can only register Interventions with this register"
            raise TypeError(msg)

        if self._check_new_intervention(intervention):

            for key, value in intervention.data.items():
                self._register_attribute(key, value)

            numeric_asset = [0] * len(self._attribute_keys)

            for key, value in intervention.data.items():
                attr_idx = self.attribute_index(key)
                value_idx = self.attribute_value_index(attr_idx, value)
                numeric_asset[attr_idx] = value_idx

            self._numeric_keys.append(numeric_asset)
            self._names[intervention.name] = numeric_asset

    def _register_attribute(self, key, value):
        """Add a new attribute and its possible value to the register (or, if
        the attribute has been seen before, add a new possible value)
        """
        if key not in self._attribute_keys:
            self._attribute_keys.append(key)
            self._attribute_possible_values.append([None])

        attr_idx = self.attribute_index(key)

        if value not in self._attribute_possible_values[attr_idx]:
            self._attribute_possible_values[attr_idx].append(value)

    def attribute_index(self, key):
        """Get the index of an attribute name
        """
        return list(self._attribute_keys).index(key)

    def attribute_value_index(self, attr_idx, value):
        """Get the index of a possible value for a given attribute index
        """
        return self._attribute_possible_values[attr_idx].index(value)

    def numeric_to_intervention(self, numeric_asset):
        """Convert the numeric representation of an asset back to Asset (with
        legible key/value data)

        Parameters
        ----------
        numeric_asset : list
            A list of integers of length `self._attribute_keys`

        Returns
        -------
        Intervention
            An :class:`Intervention` object

        Examples
        --------

        Given a (very minimal) possible state of a register:

        >>> register = AssetRegister()
        >>> register._names = [[1,1,1]]
        >>> register._attribute_keys = ["name", "capacity", "sector"]
        >>> register._attribute_possible_values = [
        ...     [None, "water_treatment_plant"],
        ...     [None, {"value": 5, "units": "ML/day"}],
        ...     [None, "water_supply"]
        ... ]

        Calling this function would piece together the asset:

        >>> asset = register.numeric_to_asset([1,1,1])
        >>> print(asset)
        Asset("water_treatment_plant", {"name": "water_treatment_plant",
        "capacity": {"units": "ML/day", "value": 5}, "sector": "water_supply"})

        """
        data = {}
        for attr_idx, value_idx in enumerate(numeric_asset):
            key = list(self._attribute_keys)[attr_idx]
            value = self._attribute_possible_values[attr_idx][value_idx]

            data[key] = value

        intervention = Intervention(data=data)

        return intervention

    def __iter__(self):
        """Iterate over the list of asset types held in the register
        """
        for asset in self._numeric_keys:
            yield self.numeric_to_intervention(asset)

    def __len__(self):
        """Returns the number of asset types stored in the register
        """
        return len(self._numeric_keys)



"""Solve the optimal planning problem for a system of systems model

"""

from pyomo.environ import (AbstractModel, Binary, Constraint,
                           NonNegativeIntegers, Objective, Param, RangeSet,
                           Set, Var, minimize, summation)
from pyomo.opt import SolverFactory


def define_basic_model(assets, availability_constraint, asset_costs):
    """Define the binary integer planning problem

    Arguments
    =========
    assets : list
        The list of assets
    availability_constraint : dict
        A dictionary of binary constraints on whether you can build
        each asset in `assets`
    asset_costs : dict
        The investment cost of each asset
    asset_value : dict
        The value function approximation of each asset

    Returns
    =======
    model : pyomo.environ.ConcreteModel
        An abstract instance of the incomplete model with no objective function
    """
    model = AbstractModel()

    number_states = 2 ** len(assets)

    model.p = Param(default=number_states,
                    doc='The number of states',
                    within=NonNegativeIntegers)

    model.I = Set(initialize=assets,
                  doc='The set of infrastructure assets `I`')

    model.b = Param(model.I,
                    initialize=availability_constraint,
                    doc='Constraint on availability of asset `i`')

    model.c = Param(model.I,
                    initialize=asset_costs,
                    doc='Cost of asset `i`')

    model.x = Var(model.I,
                  doc='Decision to invest in asset `i`',
                  domain=Binary)

    def ax_constraint_rule(model, i):
        """Implements the action constraint

        Defines which assets can be built
        """
        return model.x[i] <= model.b[i]

    model.BuildConstraint = Constraint(model.I, rule=ax_constraint_rule)

    return model


def state_vfa_model(assets, availability_constraint, asset_costs, asset_value,
                    states):
    """Define the value function approximation

    Here we assume that the value function approximation is a function
    of the state, rather than individual assets

    Unfortunately, the number of states becomes very large,
    growing exponentially in the number of assets, and so representing the
    approximate value function like this is very inefficient as soon as the
    number of assets increases above 32 (about 4 GB).

    Arguments
    =========
    assets : list
        The list of assets
    availability_constraint : dict
        A dictionary of binary constraints on whether you can build
        each asset in `assets`
    asset_costs : dict
        The investment cost of each asset
    asset_value : dict
        The value function approximation of each asset
    states : dict
        A dictionary where the keys are tuples of entries in `assets` and an
        index of states (``2 ** len(assets)``) and the value is a binary
        indicator showing the possible combinations

    Returns
    =======
    model : pyomo.environ.ConcreteModel
        A concrete instance of the model

    Notes
    =====

    :math:`\hat{v}_t^n = \min_{a_t \in A_t^n} C_t^{INV}(S^n_t, a^n_t) + \
    {V}^n_t(S^n_t, a^n_t)`

    """
    model = define_basic_model(assets, availability_constraint, asset_costs)

    model.J = RangeSet(1, model.p,
                       doc='The set of states')

    model.d = Param(model.J,
                    initialize=asset_value,
                    doc='Value of being in state `j`')

    model.y = Var(model.J,
                  doc='Decision to NOT invest in state `j`',
                  domain=Binary)

    model.e = Param(model.I, model.J,
                    initialize=states,
                    doc='The combination to match')

    def left_constraint(model, i, j):
        """Forces state `j` to be chosen if x[i] matches e[i,j]
        """
        return (model.x[i] - model.e[i, j]) <= \
            0 + (model.p + 1) * model.y[j]

    model.LeftConstraint = Constraint(model.I, model.J, rule=left_constraint)

    def right_constraint(model, i, j):
        """Forces state `j` to be chosen if x[i] matches e[i,j]
        """
        return 0 <= \
            (model.x[i] - model.e[i, j]) + (model.p + 1) * model.y[j]

    model.RightConstraint = Constraint(model.I, model.J, rule=right_constraint)

    def only_one_constraint(model):
        """Only allow investment in one state

        Right hand side of constraint is N-k (number of assets minus one).

        """
        return summation(model.y) == (model.p - 1)

    model.UniqueConstraint = Constraint(rule=only_one_constraint)

    def objective_function(model):
        """Sum of investment cost of asset `i` and value of being in state `j`
        """
        return summation(model.c, model.x) + \
            sum(model.d[j] * (1 - model.y[j]) for j in model.J)

    model.OBJ = Objective(rule=objective_function,
                          sense=minimize)

    return model


def linear_vfa_model(assets, availability_constraint, asset_costs,
                     asset_value):
    """Define the value function approximation

    Here we assume a linear relationship (in the asset)

    Arguments
    =========
    assets : list
        The list of assets
    availability_constraint : dict
        A dictionary of binary constraints on whether you can build
        each asset in `assets`
    asset_costs : dict
        The investment cost of each asset
    asset_value : dict
        The value function approximation of each asset

    Returns
    =======
    model : pyomo.environ.ConcreteModel
        A concrete instance of the model
    """

    model = define_basic_model(assets, availability_constraint, asset_costs)

    model.d = Param(model.I,
                    initialize=asset_value,
                    doc='Value of asset `i` (assumes linear relationship)')

    def obj_expression(model):
        """Total cost
        """
        return summation(model.c, model.x) + summation(model.d, model.x)

    model.OBJ = Objective(rule=obj_expression,
                          sense=minimize)

    return model


def solve_model(model, state=None):
    """Solves the model using glpk

    Passing in a `state` as a list of asset names initialises the state
    of the model, fixing those decision variables.

    Arguments
    =========
    model : pyomo.environ.ConcreteModel
        A concrete instance of the model
    state : dict, optional, default=None
        A list of assets which were installed in a previous iteration or time
        period

    returns : pyomo.core.Model.Instance
        An instance of the model populated with results
    """

    opt = SolverFactory('glpk')

    instance = model.create_instance()

    if state:
        for asset in state:
            instance.x[asset] = 1
            instance.x[asset].fixed = True
        instance.preprocess()

    results = opt.solve(instance)

    instance.solutions.load_from(results)
    instance.display()

    return instance


def feature_vfa_model(assets, availability_constraint, asset_costs,
                      feature_coefficients, asset_features):
    """Define the value function approximation

    Here we assume that the value function approximation is a function
    of the features of the state, rather than individual assets, or enumeration
    of all possible states

    Arguments
    =========
    assets : list
        The list of assets
    availability_constraint : dict
        A dictionary of binary constraints on whether you can build
        each asset in `assets`
    asset_costs : dict
        The investment cost of each asset
    features : list
        The set of features
    feature_coefficients : dict
        The regression coefficients for each feature
    asset_features : dict
        The mapping of features to assets

    Returns
    =======
    model : pyomo.environ.ConcreteModel
        A concrete instance of the model

    Notes
    =====
    The use of features decomposes the optimisation problem into several
    sub-components.  The first is to find the clusters of 'bits' in the state,
    which correctly predict minimal cost investments
    and operation of infrastructure assets.

    These clusters can be used to define a feature, adding an entry to the
    `feature` set of features and a column to the `asset_features` matrix,
    where 1 indicates that the feature includes that asset.

    Initially, feature selection could begin by regressing the results from
    a random sample of investment policies. This could highlight which
    patterns of investments seem to result in least cost systems. In addition,
    the attributes of the assets could provide a set of features which at least
    help categorise the assets (e.g. according to sector, size, location,
    and asset type).

    The binary integer problem is posed as follows:

    :math:`\min \sum_i^I c_i x_i + x_i \sum_f^{F}( e_{if} b_{f})`

    where

    - :math:`i` is an element in the set of assets :math:`I`
    - :math:`f` is an element in the set of features :math:`F`
    - :math:`c_i` is the cost of asset :math:`i`
    - :math:`x_i` is the decision to invest in asset :math:`i`
    - :math:`e_{if}` is the mapping of feature :math:`f` to asset :math:`i`
    - :math:`b_{f}` is the basis coefficient of asset :math:`i` and
      feature :math:`f`

    """
    model = define_basic_model(assets, availability_constraint, asset_costs)

    features = list(feature_coefficients.keys())

    model.F = Set(initialize=features,
                  doc='The set of basis functions')

    model.coef = Param(model.F,
                       initialize=feature_coefficients,
                       doc='The basis values')

    model.basis = Param(model.F, model.I,
                        initialize=asset_features,
                        doc='The asset-feature mapping',
                        within=Binary)

    def weight_features(model, i):
        """Computes the blend of features associated with asset `i`

        Arguments
        =========
        model : pyomo.core.AbstractModel
            An instance of the model
        i : pyomo.core.Set
            An element in the set of assets

        Returns
        =======
        expr : pyomo.core.Expression
            The blend of features of asset `i`
        """
        expr = sum(model.coef[f] * model.basis[f, i] for f in model.F)
        return expr

    model.weighted = Param(model.I,
                           initialize=weight_features,
                           doc='The weighted basis values for each asset')

    def objective_function(model):
        return summation(model.c, model.x) + summation(model.weighted, model.x)

    model.OBJ = Objective(rule=objective_function,
                          sense=minimize)

    return model


def formulate_model(asset_register, availability_constraint,
                    feature_coefficients, asset_features):

    assets = [asset.name for asset in asset_register]
    costs = [asset.data['capital cost']['value'] for asset in asset_register]
    asset_costs = dict(zip(assets, costs))

    model = feature_vfa_model(assets, availability_constraint, asset_costs,
                              feature_coefficients, asset_features)

    return model



# -*- coding: utf-8 -*-
"""This module coordinates the software components that make up the integration
framework.

"""
import logging
import operator
from collections import defaultdict

import networkx
from enum import Enum
from smif import SpaceTimeValue, StateData
from smif.convert import SpaceTimeConvertor
from smif.convert.area import RegionRegister, RegionSet
from smif.convert.interval import TimeIntervalRegister
from smif.decision import Planning
from smif.intervention import Intervention, InterventionRegister
from smif.sector_model import SectorModelBuilder

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class SosModel(object):
    """Consists of the collection of timesteps and sector models

    This is NISMOD - i.e. the system of system model which brings all of the
    sector models together. Sector models may be joined through dependencies.

    This class is populated at runtime by the :class:`SosModelBuilder` and
    called from :func:`smif.cli.run_model`.

    Attributes
    ==========
    models : dict
        This is a dictionary of :class:`smif.SectorModel`
    initial_conditions : list
        List of interventions required to set up the initial system, with any
        state attributes provided here too

    """
    def __init__(self):
        # housekeeping
        self.logger = logging.getLogger(__name__)
        self.max_iterations = 25

        # models
        self.models = {}
        self.dependency_graph = None

        # space and time
        self._timesteps = []
        self.regions = RegionRegister()
        self.intervals = TimeIntervalRegister()
        self._resolution_mapping = {'scenario': {}}

        # systems, interventions and (system) state
        self.interventions = InterventionRegister()
        self.initial_conditions = []
        self.planning = Planning([])
        self._state = defaultdict(dict)

        # scenario data and results
        self._scenario_data = {}
        self._results = defaultdict(dict)

    @property
    def resolution_mapping(self):
        """Returns the temporal and spatial mapping to an input, output or scenario parameter

        Example
        -------
        The data structure follows ``source->parameter->{temporal, spatial}``::

                {
                    'scenario': {
                        'raininess': {
                            'temporal_resolution': 'annual',
                            'spatial_resolution': 'LSOA'
                        }
                    }
                }

        """
        return self._resolution_mapping

    @resolution_mapping.setter
    def resolution_mapping(self, value):
        self._resolution_mapping = value

    @property
    def scenario_data(self):
        """Get nested dict of scenario data

        Returns
        -------
        dict
            Nested dictionary in the format ``data[year][param] =
            SpaceTimeValue(region, interval, value, unit)``
        """
        return self._scenario_data

    @property
    def results(self):
        """Get nested dict of model results

        Returns
        -------
        dict
            Nested dictionary in the format
            results[int:year][str:model][str:parameter] => list of
            SpaceTimeValues
        """
        # convert from defaultdict to plain dict
        return dict(self._results)

    def run(self):
        """Runs the system-of-system model

        0. Determine run mode

        1. Determine running order

        2. Run each sector model

        3. Return success or failure
        """
        mode = self.determine_running_mode()
        self.logger.debug("Running in %s mode", mode.name)

        if mode == RunMode.static_simulation:
            self._run_static_sos_model()
        elif mode == RunMode.sequential_simulation:
            self._run_sequential_sos_model()
        elif mode == RunMode.static_optimisation:
            self._run_static_optimisation()
        elif mode == RunMode.dynamic_optimisation:
            self._run_dynamic_optimisation()

    def _run_static_sos_model(self):
        """Runs the system-of-system model for one timeperiod

        Calls each of the sector models in the order required by the graph of
        dependencies, passing in the year for which they need to run.

        """
        run_order = self._get_model_sets_in_run_order()
        timestep = self.timesteps[0]

        for model_set in run_order:
            model_set.run(timestep)

    def _run_sequential_sos_model(self):
        """Runs the system-of-system model sequentially
        """
        run_order = self._get_model_sets_in_run_order()
        self.logger.info("Determined run order as %s", run_order)
        for timestep in self.timesteps:
            for model_set in run_order:
                model_set.run(timestep)

    def run_sector_model(self, model_name):
        """Runs the sector model

        Parameters
        ----------
        model_name : str
            The name of the model, corresponding to the folder name in the
            models subfolder of the project folder
        """
        msg = "Model '{}' does not exist. Choose from {}"
        assert model_name in self.models, \
            msg.format(model_name, self.sector_models)

        msg = "Running the %s sector model"
        self.logger.info(msg, model_name)

        sector_model = self.models[model_name]

        # Run a simulation for a single model
        for timestep in self.timesteps:
            state, results = self.run_sector_model_timestep(sector_model, timestep)
            self.set_state(sector_model, timestep, state)
            self.set_data(sector_model, timestep, results)

    def run_sector_model_timestep(self, model, timestep):
        """Run the sector model for a specific timestep

        Parameters
        ----------
        model: :class:`smif.sector_model.SectorModel`
            The instance of the sector model wrapper to run
        timestep: int
            The year for which to run the model

        """
        decisions = self.get_decisions(model, timestep)
        state = self.get_state(model, timestep)
        data = self.get_data(model, timestep)

        state, results = model.simulate(decisions, state, data)
        self.logger.debug("Results from %s model:\n %s", model.name, results)
        return state, results

    def get_decisions(self, model, timestep):
        """Gets the interventions that correspond to the decisions

        Parameters
        ----------
        model: :class:`smif.sector_model.SectorModel`
            The instance of the sector model wrapper to run
        timestep: int
            The current model year
        """
        self.logger.debug("Finding decisions for %i", timestep)
        current_decisions = []
        for decision in self.planning.planned_interventions:
            if decision['build_date'] <= timestep:
                name = decision['name']
                if name in model.intervention_names:
                    msg = "Adding decision '%s' to instruction list"
                    self.logger.debug(msg, name)
                    intervention = self.interventions.get_intervention(name)
                    current_decisions.append(intervention)
        # for decision in self.planning.get_rule_based_interventions(timestep):
        #   current_decisions.append(intervention)
        # for decision in self.planning.get_optimised_interventions(timestep):
        #   current_decisions.append(intervention)

        return current_decisions

    def get_state(self, model, timestep):
        """Gets the state to pass to SectorModel.simulate
        """
        if model.name not in self._state[timestep]:
            self.logger.warning("Found no state for %s in timestep %s", model.name, timestep)
            return []
        return self._state[timestep][model.name]

    def set_state(self, model, from_timestep, state):
        """Sets state output from model ready for next timestep
        """
        for_timestep = self.timestep_after(from_timestep)
        self._state[for_timestep][model.name] = state

    def get_data(self, model, timestep):
        """Gets the data in the required format to pass to the simulate method

        Returns
        -------
        dict
            A nested dictionary of the format:
            ``data[parameter][region][time_interval] = {value, units}``

        Notes
        -----
        Note that the timestep is `not` passed to the SectorModel in the
        nested data dictionary.
        The current timestep is available in ``data['timestep']``.

        """
        new_data = {}
        for dependency in model.inputs.parameters:
            name = dependency.name
            provider = self.outputs[name]

            for source in provider:

                self.logger.debug("Getting '%s' dependency data for '%s' from '%s'",
                                  name, model.name, source)

                if source == 'scenario':
                    from_data = self.scenario_data[timestep][name]
                    scenario_map = self.resolution_mapping['scenario']
                    from_spatial_resolution = scenario_map[name]['spatial_resolution']
                    from_temporal_resolution = scenario_map[name]['temporal_resolution']
                    self.logger.debug("Found data: %s", from_data)

                elif source in self.models:
                    source_model = self.models[source]
                    # get latest set of results from list
                    from_data = self.results[timestep][source][name]
                    from_spatial_resolution = source_model.outputs.get_spatial_res(name)
                    from_temporal_resolution = source_model.outputs.get_temporal_res(name)
                    self.logger.debug("Found data: %s", from_data)

                else:
                    msg = "The data source for dependency %s was not found"
                    raise ValueError(msg, name)

                to_spatial_resolution = dependency.spatial_resolution
                to_temporal_resolution = dependency.temporal_resolution
                msg = "Converting from spatial resolution '%s' and  temporal resolution '%s'"
                self.logger.debug(msg, from_spatial_resolution, from_temporal_resolution)
                msg = "Converting to spatial resolution '%s' and  temporal resolution '%s'"
                self.logger.debug(msg, to_spatial_resolution, to_temporal_resolution)

                if name not in new_data:
                    new_data[name] = self._convert_data(from_data,
                                                        to_spatial_resolution,
                                                        to_temporal_resolution,
                                                        from_spatial_resolution,
                                                        from_temporal_resolution)
                else:
                    to_add = self._convert_data(from_data,
                                                to_spatial_resolution,
                                                to_temporal_resolution,
                                                from_spatial_resolution,
                                                from_temporal_resolution)
                    new_data[name] = SosModel.add_data_series(new_data[name], to_add)

        new_data['timestep'] = timestep
        return new_data

    def set_data(self, model, timestep, results):
        """Sets results output from model as data available to other/future models

        Stores only latest estimated results (i.e. not holding on to iterations
        here while trying to solve interdependencies)
        """
        self._results[timestep][model.name] = results

    def _convert_data(self, data, to_spatial_resolution,
                      to_temporal_resolution, from_spatial_resolution,
                      from_temporal_resolution):
        """Given a model, check required parameters, pick data from scenario
        for the given timestep

        Parameters
        ----------
        timestep: int
            The year for which to get scenario data
        dependency: :class:`smif.SpaceTimeValue`

        Returns
        -------
        list
            A list of :class:`SpaceTimeValue`

        """
        convertor = SpaceTimeConvertor(data,
                                       from_spatial_resolution,
                                       to_spatial_resolution,
                                       from_temporal_resolution,
                                       to_temporal_resolution,
                                       self.regions,
                                       self.intervals)
        return convertor.convert()

    @staticmethod
    def add_data_series(list_a, list_b):
        """Given two lists of SpaceTimeValues of identical spatial and temporal
        resolution, return a single list with matching values added together.

        Notes
        -----
        Assumes a data series is not sparse, i.e. has a value for every
        region/interval combination
        """
        list_a.sort(key=operator.attrgetter('region', 'interval'))
        list_b.sort(key=operator.attrgetter('region', 'interval'))

        return [a + b for a, b in zip(list_a, list_b)]

    def _run_static_optimisation(self):
        """Runs the system-of-systems model in a static optimisation format
        """
        raise NotImplementedError

    def _run_dynamic_optimisation(self):
        """Runs the system-of-system models in a dynamic optimisation format
        """
        raise NotImplementedError

    def _get_model_sets_in_run_order(self):
        """Returns a list of :class:`ModelSet` in a runnable order.

        If a set contains more than one model, there is an interdependency and
        and we attempt to run the models to convergence.
        """
        if networkx.is_directed_acyclic_graph(self.dependency_graph):
            # topological sort gives a single list from directed graph, currently
            # ignoring opportunities to run independent models in parallel
            run_order = networkx.topological_sort(self.dependency_graph, reverse=True)

            # turn into a list of sets for consistency with the below
            ordered_sets = [
                ModelSet(
                    {self.models[model_name]},
                    self
                )
                for model_name in run_order
            ]

        else:
            # contract the strongly connected components (subgraphs which
            # contain cycles) into single nodes, producing the 'condensation'
            # of the graph, where each node maps to one or more sector models
            condensation = networkx.condensation(self.dependency_graph)

            # topological sort of the condensation gives an ordering of the
            # contracted nodes, whose 'members' attribute refers back to the
            # original dependency graph
            ordered_sets = [
                ModelSet(
                    {
                        self.models[model_name]
                        for model_name in condensation.node[node_id]['members']
                    },
                    self
                )
                for node_id in networkx.topological_sort(condensation, reverse=True)
            ]

        return ordered_sets

    def determine_running_mode(self):
        """Determines from the config in what mode to run the model

        Returns
        =======
        :class:`RunMode`
            The mode in which to run the model
        """

        number_of_timesteps = len(self._timesteps)

        if number_of_timesteps > 1:
            # Run a sequential simulation
            mode = RunMode.sequential_simulation

        elif number_of_timesteps == 0:
            raise ValueError("No timesteps have been specified")

        else:
            # Run a single simulation
            mode = RunMode.static_simulation

        return mode

    @property
    def timesteps(self):
        """Returns the list of timesteps

        Returns
        =======
        list
            A list of timesteps, distinct and sorted in ascending order
        """
        return self._timesteps

    @timesteps.setter
    def timesteps(self, value):
        self._timesteps = sorted(list(set(value)))

    def timestep_before(self, timestep):
        """Returns the timestep previous to a given timestep, or None
        """
        if timestep not in self.timesteps or timestep == self.timesteps[0]:
            return None
        else:
            index = self.timesteps.index(timestep)
            return self.timesteps[index - 1]

    def timestep_after(self, timestep):
        """Returns the timestep after a given timestep, or None
        """
        if timestep not in self.timesteps or timestep == self.timesteps[-1]:
            return None
        else:
            index = self.timesteps.index(timestep)
            return self.timesteps[index + 1]

    @property
    def intervention_names(self):
        """Names (id-like keys) of all known asset type
        """
        return [intervention.name for intervention in self.interventions]

    @property
    def sector_models(self):
        """The list of sector model names

        Returns
        =======
        list
            A list of sector model names
        """
        return list(self.models.keys())

    @property
    def inputs(self):
        """A dictionary of model names associated with an inputs

        Returns
        -------
        dict
            Keys are parameter names, value is a list of sector model names
        """
        parameter_model_map = defaultdict(list)
        for model_name, model in self.models.items():
            for dep in model.inputs.parameters:
                parameter_model_map[dep.name].append(model_name)
        return parameter_model_map

    @property
    def outputs(self):
        """Model names associated with model outputs & scenarios

        Returns
        -------
        dict
            Keys are parameter names, value is a list of sector model names
        """
        parameter_model_map = defaultdict(list)
        for model_name, model_data in self.models.items():
            for output in model_data.outputs.parameters:
                parameter_model_map[output.name].append(model_name)

        for name in self.resolution_mapping['scenario'].keys():
            parameter_model_map[name].append('scenario')
        return parameter_model_map


class ModelSet(object):
    """Wraps a set of interdependent models

    Given a directed graph of dependencies between models, any cyclic
    dependencies are contained within the strongly-connected components of the
    graph.

    A ModelSet corresponds to the set of models within a single strongly-
    connected component. If this is a set of one model, it can simply be run
    deterministically. Otherwise, this class provides the machinery necessary
    to find a solution to each of the interdependent models.

    The current implementation first estimates the outputs for each model in the
    set, guaranteeing that each model will then be able to run, then begins
    iterating, running every model in the set at each iteration, monitoring the
    model outputs over the iterations, and stopping at timeout, divergence or
    convergence.

    Notes
    -----
    This calls back into :class:`SosModel` quite extensively for state, data,
    decisions, regions and intervals.

    """
    def __init__(self, models, sos_model):
        self.logger = logging.getLogger(__name__)
        self._models = models
        self._model_names = {model.name for model in models}
        self._sos_model = sos_model
        self.iterated_results = {}

    def run(self, timestep):
        """Runs a set of one or more models
        """
        if len(self._models) == 1:
            # Short-circuit if the set contains a single model - this
            # can be run deterministically
            model = list(self._models)[0]
            logging.debug("Running %s for %d", model.name, timestep)
            state, results = self._sos_model.run_sector_model_timestep(model, timestep)
            self._sos_model.set_state(model, timestep, state)
            self._sos_model.set_data(model, timestep, results)
        else:
            # Start by running all models in set with best guess
            # - zeroes
            # - last year's inputs
            self.iterated_results = {}
            for model in self._models:
                results = self.guess_results(model, timestep)
                self._sos_model.set_data(model, timestep, results)
                self.iterated_results[model.name] = [results]

            # - keep track of intermediate results (iterations within the timestep)
            # - stop iterating according to near-equality condition
            for i in range(self._sos_model.max_iterations):
                if self.converged(timestep):
                    break
                else:
                    self.logger.debug("Iteration %s, model set %s", i, self._model_names)
                    for model in self._models:
                        state, results = self._sos_model.run_sector_model_timestep(
                            model, timestep)
                        self._sos_model.set_state(model, timestep, state)
                        self._sos_model.set_data(model, timestep, results)
                        self.iterated_results[model.name].append(results)
            else:
                raise TimeoutError("Model evaluation exceeded max iterations")

    def guess_results(self, model, timestep):
        """Dependency-free guess at a model's result set.

        Initially, guess zeroes, or the previous timestep's results.
        """
        timestep_before = self._sos_model.timestep_before(timestep)
        if timestep_before is not None:
            # last iteration of previous timestep results
            results = self._sos_model.results[timestep_before][model.name]
        else:
            # generate zero-values for each parameter/region/interval combination
            results = {}
            for output in model.outputs.parameters:
                output_results = []
                regions = self._sos_model.regions.get_regions_in_set(
                    output.spatial_resolution)
                intervals = self._sos_model.intervals.get_intervals_in_set(
                    output.temporal_resolution)
                for region in regions:
                    region_name = region.name
                    for interval_name in intervals.keys():
                        output_results.append(
                            SpaceTimeValue(
                                region_name,
                                interval_name,
                                0,
                                "unknown"
                            )
                        )
                results[output.name] = output_results
        return results

    def converged(self, timestep):
        """Check whether the results of a set of models have converged.

        Returns
        -------
        converged: bool
            True if the results have converged to within a tolerance

        Raises
        ------
        DiverganceError
            If the results appear to be diverging
        """
        model_set_results = [
            self.iterated_results[model_name]
            for model_name in self._model_names
        ]

        if any([len(results) < 2 for results in model_set_results]):
            # must have at least two result sets per model to assess convergence
            return False

        if all([results[-1] == results[-2] for results in model_set_results]):
            # if all most recent are exactly equal to penultimate, must have converged
            return True

        return False


class SosModelBuilder(object):
    """Constructs a system-of-systems model

    Builds a :class:`SosModel`.

    Examples
    --------
    Call :py:meth:`SosModelBuilder.construct` to populate
    a :class:`SosModel` object and :py:meth:`SosModelBuilder.finish`
    to return the validated and dependency-checked system-of-systems model.

    >>> builder = SosModelBuilder()
    >>> builder.construct(config_data)
    >>> sos_model = builder.finish()

    """
    def __init__(self):
        self.sos_model = SosModel()

        self.logger = logging.getLogger(__name__)

    def construct(self, config_data):
        """Set up the whole SosModel

        Parameters
        ----------
        config_data : dict
            A valid system-of-systems model configuration dictionary
        """
        model_list = config_data['sector_model_data']

        self.add_timesteps(config_data['timesteps'])
        self.set_max_iterations(config_data)

        self.load_region_sets(config_data['region_sets'])
        self.load_interval_sets(config_data['interval_sets'])

        self.load_models(model_list)
        self.add_planning(config_data['planning'])
        self.add_resolution_mapping(config_data['resolution_mapping'])
        self.add_scenario_data(config_data['scenario_data'])
        self.logger.debug(config_data['scenario_data'])

    def add_timesteps(self, timesteps):
        """Set the timesteps of the system-of-systems model

        Parameters
        ----------
        timesteps : list
            A list of timesteps
        """
        self.logger.info("Adding timesteps")
        self.sos_model.timesteps = timesteps

    def set_max_iterations(self, config_data):
        if 'max_iterations' in config_data and config_data['max_iterations'] is not None:
            self.sos_model.max_iterations = config_data['max_iterations']

    def add_resolution_mapping(self, resolution_mapping):
        """

        Parameters
        ----------
        resolution_mapping: dict
            A dictionary containing information on the spatial and temporal
            resoultion of scenario data

        Example
        -------
        The data structure follows ``source->parameter->{temporal, spatial}``::

                {'scenario': {
                 'raininess': {'temporal_resolution': 'annual',
                               'spatial_resolution': 'LSOA'}}}

        """
        self.sos_model.resolution_mapping = resolution_mapping

    def load_region_sets(self, region_sets):
        """Loads the region sets into the system-of-system model

        Parameters
        ----------
        region_sets: list
            A dict, where key is the name of the region set, and the value
            the data
        """
        assert isinstance(region_sets, dict)

        region_set_definitions = region_sets.items()
        if len(region_set_definitions) == 0:
            msg = "No region sets have been defined"
            self.logger.warning(msg)
        for name, data in region_set_definitions:
            msg = "Region set data is not a list"
            assert isinstance(data, list), msg
            self.sos_model.regions.register(RegionSet(name, data))

    def load_interval_sets(self, interval_sets):
        """Loads the time-interval sets into the system-of-system model

        Parameters
        ----------
        interval_sets: list
            A dict, where key is the name of the interval set, and the value
            the data
        """
        interval_set_definitions = interval_sets.items()
        if len(interval_set_definitions) == 0:
            msg = "No interval sets have been defined"
            self.logger.warning(msg)

        for name, data in interval_set_definitions:
            self.sos_model.intervals.register(data, name)

    def load_models(self, model_data_list):
        """Loads the sector models into the system-of-systems model

        Parameters
        ----------
        model_data_list : list
            A list of sector model config/data
        assets : list
            A list of assets to pass to the sector model

        """
        self.logger.info("Loading models")
        for model_data in model_data_list:
            model = self._build_model(model_data)
            self.add_model(model)
            self.add_model_data(model, model_data)

    @staticmethod
    def _build_model(model_data):
        builder = SectorModelBuilder(model_data['name'])
        builder.load_model(model_data['path'], model_data['classname'])
        builder.create_initial_system(model_data['initial_conditions'])
        builder.add_inputs(model_data['inputs'])
        builder.add_outputs(model_data['outputs'])
        builder.add_interventions(model_data['interventions'])
        return builder.finish()

    def add_model(self, model):
        """Adds a sector model to the system-of-systems model

        Parameters
        ----------
        model : :class:`smif.sector_model.SectorModel`
            A sector model wrapper

        """
        self.logger.info("Loading model: %s", model.name)
        self.sos_model.models[model.name] = model

    def add_model_data(self, model, model_data):
        """Adds sector model data to the system-of-systems model which is
        convenient to have available at the higher level.
        """
        self.add_initial_conditions(model.name, model_data['initial_conditions'])
        self.add_interventions(model.name, model_data['interventions'])

    def add_interventions(self, model_name, interventions):
        """Adds interventions for a model
        """
        for intervention in interventions:
            intervention_object = Intervention(sector=model_name,
                                               data=intervention)
            msg = "Adding %s from %s to SosModel InterventionRegister"
            identifier = intervention_object.name
            self.logger.debug(msg, identifier, model_name)
            self.sos_model.interventions.register(intervention_object)

    def add_initial_conditions(self, model_name, initial_conditions):
        """Adds initial conditions (state) for a model
        """
        timestep = self.sos_model.timesteps[0]
        state_data = filter(
            lambda d: len(d.data) > 0,
            [self.intervention_state_from_data(datum) for datum in initial_conditions]
        )
        self.sos_model._state[timestep][model_name] = list(state_data)

    @staticmethod
    def intervention_state_from_data(intervention_data):
        """Unpack an intervention from the initial system to extract StateData
        """
        target = None
        data = {}
        for key, value in intervention_data.items():
            if key == "name":
                target = value

            if isinstance(value, dict) and "is_state" in value and value["is_state"]:
                del value["is_state"]
                data[key] = value

        return StateData(target, data)

    def add_planning(self, planning):
        """Loads the planning logic into the system of systems model

        Pre-specified planning interventions are defined at the sector-model
        level, read in through the SectorModel class, but populate the
        intervention register in the controller.

        Parameters
        ----------
        planning : list
            A list of planning instructions

        """
        self.logger.info("Adding planning")
        self.sos_model.planning = Planning(planning)

    def add_scenario_data(self, data):
        """Load the scenario data into the system of systems model

        Expect a dictionary, where each key maps a parameter
        name to a list of data, each observation with:

        - timestep
        - value
        - units
        - region (must use a region id from scenario regions)
        - interval (must use an id from scenario time intervals)

        Add a dictionary of list of :class:`smif.SpaceTimeValue` named
        tuples,
        for ease of iteration::

                data[year][param] = SpaceTimeValue(region, interval, value, units)

        Default region: "national"
        Default interval: "annual"
        """
        self.logger.info("Adding scenario data")
        nested = {}

        for param, observations in data.items():
            if param not in self.sos_model.resolution_mapping['scenario']:
                raise ValueError("Parameter {} not registered in resolution mapping {}".format(
                    param,
                    self.sos_model.resolution_mapping))
            resolution_sets = self.sos_model.resolution_mapping['scenario'][param]

            interval_set_name = resolution_sets['temporal_resolution']
            interval_set = self.sos_model.intervals.get_intervals_in_set(interval_set_name)
            interval_names = [interval.name for key, interval in interval_set.items()]

            region_set_name = resolution_sets['spatial_resolution']
            region_set = self.sos_model.regions.get_regions_in_set(region_set_name)
            region_names = [region.name for region in region_set]

            for obs in observations:
                if 'year' not in obs:
                    raise ValueError("Scenario data item missing year: {}".format(obs))
                year = obs['year']
                if year not in nested:
                    nested[year] = {}

                region = obs['region']
                if region not in region_names:
                    raise ValueError(
                        "Region {} not defined in set {} for parameter {}".format(
                            region,
                            region_set_name,
                            param))

                interval = obs['interval']
                if interval not in interval_names:
                    raise ValueError(
                        "Interval {} not defined in set {} for parameter {}".format(
                            interval,
                            interval_set_name,
                            param))

                entry = SpaceTimeValue(
                    region,
                    interval,
                    obs['value'],
                    obs['units']
                )
                if param not in nested[year]:
                    nested[year][param] = [entry]
                else:
                    nested[year][param].append(entry)

        self.sos_model._scenario_data = nested

    def _check_planning_interventions_exist(self):
        """Check existence of all the interventions in the pre-specifed planning

        """
        model = self.sos_model
        names = model.intervention_names
        for planning_name in model.planning.names:
            msg = "Intervention '{}' in planning file not found in interventions"
            assert planning_name in names, msg.format(planning_name)

    def _check_planning_timeperiods_exist(self):
        """Check existence of all the timeperiods in the pre-specified planning
        """
        model = self.sos_model
        model_timeperiods = model.timesteps
        for timeperiod in model.planning.timeperiods:
            msg = "Timeperiod '{}' in planning file not found model config"
            assert timeperiod in model_timeperiods, msg.format(timeperiod)

    def _validate(self):
        """Validates the sos model
        """
        self._check_planning_interventions_exist()
        self._check_planning_timeperiods_exist()
        self._check_dependencies()
        self._check_region_interval_sets()

    def _check_region_interval_sets(self):
        """For each model, check for the interval and region sets referenced

        Each model references interval and region sets in the configuration
        of inputs and outputs.
        """
        available_intervals = self.sos_model.intervals.interval_set_names
        msg = "Available time interval sets in SosModel: %s"
        self.logger.debug(msg, available_intervals)
        available_regions = self.sos_model.regions.region_set_names
        msg = "Available region sets in SosModel: %s"
        self.logger.debug(msg, available_regions)

        for model_name, model in self.sos_model.models.items():
            exp_regions = []
            exp_intervals = []
            exp_regions.extend(model.inputs.spatial_resolutions)
            exp_regions.extend(model.outputs.spatial_resolutions)
            exp_intervals.extend(model.inputs.temporal_resolutions)
            exp_intervals.extend(model.outputs.temporal_resolutions)

            for region in exp_regions:
                if region not in available_regions:
                    msg = "Region set '%s' not specified but is required " + \
                          "for model '$s'"
                    raise ValueError(msg, region, model_name)
            for interval in exp_intervals:
                if interval not in available_intervals:
                    msg = "Interval set '%s' not specified but is required " + \
                          "for model '$s'"
                    raise ValueError(msg, interval, model_name)

    def _check_dependencies(self):
        """For each model, compare dependency list of from_models
        against list of available models
        """
        dependency_graph = networkx.DiGraph()
        models_available = self.sos_model.sector_models
        dependency_graph.add_nodes_from(models_available)

        for model_name, model in self.sos_model.models.items():
            for dep in model.inputs.parameters:
                providers = self.sos_model.outputs[dep.name]
                msg = "Dependency '%s' provided by '%s'"
                self.logger.debug(msg, dep.name, providers)

                if len(providers) == 0:
                    # report missing dependency type
                    msg = "Missing dependency: {} depends on {}, " + \
                        "which is not supplied."
                    raise AssertionError(msg.format(model_name, dep.name))

                for source in providers:
                    if source == 'scenario':
                        continue

                    dependency_graph.add_edge(model_name, source)

        self.sos_model.dependency_graph = dependency_graph

    def finish(self):
        """Returns a configured system-of-systems model ready for operation

        Includes validation steps, e.g. to check dependencies
        """
        self._validate()
        return self.sos_model


class RunMode(Enum):
    """Enumerates the operating modes of a SoS model
    """
    static_simulation = 0
    sequential_simulation = 1
    static_optimisation = 2
    dynamic_optimisation = 3



# -*- coding: utf-8 -*-

from pytest import fixture, raises
from smif import SpaceTimeValue
from smif.decision import Planning
from smif.sector_model import SectorModel
from smif.sos_model import ModelSet, SosModel, SosModelBuilder

from .fixtures.water_supply import WaterSupplySectorModel


@fixture(scope='function')
def get_sos_model_only_scenario_dependencies(setup_region_data):
    builder = SosModelBuilder()
    builder.add_timesteps([2010, 2011, 2012])
    builder.load_region_sets({'LSOA': setup_region_data['features']})
    interval_data = [
        {
            'id': 1,
            'start': 'P0Y',
            'end': 'P1Y'
        }
    ]
    builder.load_interval_sets({'annual': interval_data})
    builder.add_resolution_mapping({
        'scenario': {
            'raininess': {
                'temporal_resolution': 'annual',
                'spatial_resolution': 'LSOA'
            }
        }
    })
    builder.add_scenario_data({
        "raininess": [
            {
                'year': 2010,
                'value': 3,
                'units': 'ml',
                'region': 'oxford',
                'interval': 1
            },
            {
                'year': 2011,
                'value': 5,
                'units': 'ml',
                'region': 'oxford',
                'interval': 1
            },
            {
                'year': 2012,
                'value': 1,
                'units': 'ml',
                'region': 'oxford',
                'interval': 1
            }
        ]
    })

    ws = WaterSupplySectorModel()
    ws.name = 'water_supply'
    ws.inputs = [
        {
            'name': 'raininess',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]
    ws.outputs = [
        {
            'name': 'cost',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        },
        {
            'name': 'water',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]
    ws.interventions = [
        {"name": "water_asset_a", "location": "oxford"},
        {"name": "water_asset_b", "location": "oxford"},
        {"name": "water_asset_c", "location": "oxford"},
    ]
    ws_data = {
        'name': ws.name,
        'initial_conditions': [],
        'inputs': ws.inputs,
        'outputs': ws.outputs,
        'interventions': ws.interventions,
    }

    builder.add_model(ws)
    builder.add_model_data(ws, ws_data)

    ws2 = WaterSupplySectorModel()
    ws2.name = 'water_supply_2'
    ws2.inputs = [
        {
            'name': 'raininess',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]

    builder.add_model(ws2)

    return builder.finish()


@fixture(scope='function')
def get_sos_model_with_model_dependency(setup_region_data):
    builder = SosModelBuilder()
    builder.add_timesteps([2010])
    interval_data = [{'id': 1, 'start': 'P0Y', 'end': 'P1Y'}]
    builder.load_interval_sets({'annual': interval_data})
    builder.load_region_sets({'LSOA': setup_region_data['features']})
    builder.add_resolution_mapping({
        'scenario': {
            'raininess': {
                'temporal_resolution': 'annual',
                'spatial_resolution': 'LSOA'
            }
        }
    })
    builder.add_scenario_data({
        "raininess": [
            {
                'year': 2010,
                'value': 3,
                'units': 'ml',
                'region': 'oxford',
                'interval': 1
            },
            {
                'year': 2011,
                'value': 5,
                'units': 'ml',
                'region': 'oxford',
                'interval': 1
            },
            {
                'year': 2012,
                'value': 1,
                'units': 'ml',
                'region': 'oxford',
                'interval': 1
            }
        ]
    })

    ws = WaterSupplySectorModel()
    ws.name = 'water_supply'
    ws.inputs = [
        {
            'name': 'raininess',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]

    ws.outputs = [
        {
            'name': 'water',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        },
        {
            'name': 'cost',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]
    ws.interventions = [
        {"name": "water_asset_a", "location": "oxford"},
        {"name": "water_asset_b", "location": "oxford"},
        {"name": "water_asset_c", "location": "oxford"}
    ]
    builder.add_model(ws)

    ws2 = WaterSupplySectorModel()
    ws2.name = 'water_supply_2'
    ws2.inputs = [
        {
            'name': 'water',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]

    builder.add_model(ws2)

    return builder.finish()


@fixture(scope='function')
def get_sos_model_with_summed_dependency(setup_region_data):
    builder = SosModelBuilder()
    builder.add_timesteps([2010])
    builder.load_region_sets({'LSOA': setup_region_data['features']})
    interval_data = [{'id': 1, 'start': 'P0Y', 'end': 'P1Y'}]
    builder.load_interval_sets({'annual': interval_data})
    builder.add_resolution_mapping({
        'scenario': {
            'raininess': {
                'temporal_resolution': 'annual',
                'spatial_resolution': 'LSOA'
            }
        }
    })
    builder.add_scenario_data({
        "raininess": [
            {
                'year': 2010,
                'value': 3,
                'units': 'ml',
                'region': 'oxford',
                'interval': 1
            }
        ]
    })

    ws = WaterSupplySectorModel()
    ws.name = 'water_supply'
    ws.inputs = [
        {
            'name': 'raininess',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]
    ws.outputs = [
        {
            'name': 'water',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]
    builder.add_model(ws)

    ws2 = WaterSupplySectorModel()
    ws2.name = 'water_supply_2'
    ws2.inputs = [
        {
            'name': 'raininess',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]
    ws2.outputs = [
        {
            'name': 'water',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]
    builder.add_model(ws2)

    ws3 = WaterSupplySectorModel()
    ws3.inputs = [
        {
            'name': 'water',
            'spatial_resolution': 'LSOA',
            'temporal_resolution': 'annual'
        }
    ]
    ws3.name = 'water_supply_3'
    builder.add_model(ws3)

    return builder.finish()


class TestSosModel():

    def test_run_static(self, get_sos_model_only_scenario_dependencies):
        sos_model = get_sos_model_only_scenario_dependencies
        sos_model.run()

    def test_run_no_timesteps(self, get_sos_model_only_scenario_dependencies):
        sos_model = get_sos_model_only_scenario_dependencies
        sos_model.timesteps = []

        with raises(ValueError) as ex:
            sos_model.run()
        assert "No timesteps" in str(ex.value)

    def test_set_timesteps(self):
        sos_model = SosModel()
        sos_model.timesteps = [2010, 2011, 2012]
        assert sos_model.timesteps == [2010, 2011, 2012]

        # remove duplicates
        sos_model.timesteps = [2010, 2011, 2012, 2012, 2012]
        assert sos_model.timesteps == [2010, 2011, 2012]

        # sort order
        sos_model.timesteps = [2011, 2012, 2010]
        assert sos_model.timesteps == [2010, 2011, 2012]

    def test_timestep_before(self):
        sos_model = SosModel()
        sos_model.timesteps = [2010, 2011, 2012]
        assert sos_model.timestep_before(2010) is None
        assert sos_model.timestep_before(2011) == 2010
        assert sos_model.timestep_before(2012) == 2011
        assert sos_model.timestep_before(2013) is None

    def test_timestep_after(self):
        sos_model = SosModel()
        sos_model.timesteps = [2010, 2011, 2012]
        assert sos_model.timestep_after(2010) == 2011
        assert sos_model.timestep_after(2011) == 2012
        assert sos_model.timestep_after(2012) is None
        assert sos_model.timestep_after(2013) is None

    def test_guess_outputs_zero(self, get_sos_model_only_scenario_dependencies):
        """If no previous timestep has results, guess outputs as zero
        """
        sos_model = get_sos_model_only_scenario_dependencies
        ws_model = sos_model.models['water_supply']
        model_set = ModelSet(
            {ws_model},
            sos_model
        )

        results = model_set.guess_results(ws_model, 2010)
        expected = {
            "cost": [
                SpaceTimeValue("oxford", 1, 0, "unknown")
            ],
            "water": [
                SpaceTimeValue("oxford", 1, 0, "unknown")
            ]
        }
        assert results == expected

    def test_guess_outputs_last_year(self, get_sos_model_only_scenario_dependencies):
        """If a previous timestep has results, guess outputs as identical
        """
        sos_model = get_sos_model_only_scenario_dependencies
        ws_model = sos_model.models['water_supply']
        model_set = ModelSet(
            {ws_model},
            sos_model
        )

        expected = {
            "cost": [
                SpaceTimeValue("oxford", 1, 3.14, "unknown")
            ],
            "water": [
                SpaceTimeValue("oxford", 1, 2.71, "unknown")
            ]
        }

        # set up data as though from previous timestep simulation
        year_before = sos_model.timestep_before(2011)
        assert year_before == 2010
        sos_model._results[year_before]['water_supply'] = expected

        results = model_set.guess_results(ws_model, 2011)
        assert results == expected

    def test_converged_first_iteration(self, get_sos_model_only_scenario_dependencies):
        """Should not report convergence after a single iteration
        """
        sos_model = get_sos_model_only_scenario_dependencies
        ws_model = sos_model.models['water_supply']
        model_set = ModelSet(
            {ws_model},
            sos_model
        )

        results = model_set.guess_results(ws_model, 2010)
        model_set.iterated_results[ws_model.name] = [results]

        assert not model_set.converged(2010)

    def test_converged_two_identical(self, get_sos_model_only_scenario_dependencies):
        """Should report converged if the last two output sets are identical
        """
        sos_model = get_sos_model_only_scenario_dependencies
        ws_model = sos_model.models['water_supply']
        model_set = ModelSet(
            {ws_model},
            sos_model
        )

        results = model_set.guess_results(ws_model, 2010)
        model_set.iterated_results = {
            "water_supply": [results, results]
        }

        assert model_set.converged(2010)

    def test_run_sequential(self, get_sos_model_only_scenario_dependencies):
        sos_model = get_sos_model_only_scenario_dependencies
        sos_model.timesteps = [2010, 2011, 2012]
        sos_model.run()

    def test_run_single_sector(self, get_sos_model_only_scenario_dependencies):
        sos_model = get_sos_model_only_scenario_dependencies
        sos_model.run_sector_model('water_supply')

    def test_run_missing_sector(self, get_sos_model_only_scenario_dependencies):
        sos_model = get_sos_model_only_scenario_dependencies

        with raises(AssertionError) as ex:
            sos_model.run_sector_model('impossible')
        assert "Model 'impossible' does not exist" in str(ex.value)

    def test_run_with_planning(self, get_sos_model_only_scenario_dependencies):
        sos_model = get_sos_model_only_scenario_dependencies
        planning_data = [
            {
                'name': 'water_asset_a',
                'build_date': 2010,
            },
            {
                'name': 'energy_asset_a',
                'build_date': 2011,
            }
        ]
        planning = Planning(planning_data)
        sos_model.planning = planning

        model = sos_model.models['water_supply']
        actual = sos_model.get_decisions(model, 2010)
        assert actual[0].name == 'water_asset_a'
        assert actual[0].location == 'oxford'

        _, results = sos_model.run_sector_model_timestep(model, 2010)
        actual = results
        expected = {'cost': 2.528, 'water': 2}
        for key, value in expected.items():
            assert actual[key][0].value == value

        _, results = sos_model.run_sector_model_timestep(model, 2011)
        actual = results
        expected = {'cost': 2.528, 'water': 2}
        for key, value in expected.items():
            assert actual[key][0].value == value

    def test_add_data_series(self):
        """Expect addition to work on lists of SpaceTimeValues
        """
        list_a = [
            SpaceTimeValue('a', '1', 1, 'days'),
            SpaceTimeValue('a', '2', 2, 'days'),
            SpaceTimeValue('b', '1', 3, 'days'),
            SpaceTimeValue('b', '2', 4, 'days'),
        ]

        list_b = [
            SpaceTimeValue('b', '2', 14, 'days'),
            SpaceTimeValue('b', '1', 13, 'days'),
            SpaceTimeValue('a', '2', 12, 'days'),
            SpaceTimeValue('a', '1', 11, 'days'),
        ]

        expected = [
            SpaceTimeValue('a', '1', 12, 'days'),
            SpaceTimeValue('a', '2', 14, 'days'),
            SpaceTimeValue('b', '1', 16, 'days'),
            SpaceTimeValue('b', '2', 18, 'days'),
        ]
        actual = SosModel.add_data_series(list_a, list_b)
        assert actual == expected

    def test_dependency_aggregation(self, get_sos_model_with_summed_dependency):
        sos_model = get_sos_model_with_summed_dependency
        sos_model.run()


@fixture(scope='function')
def get_config_data(setup_project_folder, setup_region_data):
    path = setup_project_folder
    water_supply_wrapper_path = str(
        path.join(
            'models', 'water_supply', '__init__.py'
        )
    )
    return {
        "timesteps": [2010, 2011, 2012],
        "sector_model_data": [
            {
                "name": "water_supply",
                "path": water_supply_wrapper_path,
                "classname": "WaterSupplySectorModel",
                "inputs": [],
                "outputs": [],
                "initial_conditions": [],
                "interventions": []
            }
        ],
        "planning": [],
        "scenario_data": {
            'raininess': [
                {
                    'year': 2010,
                    'value': 3,
                    'units': 'ml',
                    'region': 'oxford',
                    'interval': 1
                }
            ]
        },
        "region_sets": {'LSOA': setup_region_data['features']},
        "interval_sets": {
            'annual': [
                {
                    'id': 1,
                    'start': 'P0Y',
                    'end': 'P1Y'
                }
            ]
        },
        "resolution_mapping": {
            'scenario': {
                'raininess': {
                    'temporal_resolution': 'annual',
                    'spatial_resolution': 'LSOA'
                }
            }
        }
    }


class TestSosModelBuilder():

    def test_builder(self, setup_project_folder):

        builder = SosModelBuilder()

        builder.add_timesteps([2010, 2011, 2012])

        model = WaterSupplySectorModel()
        model.name = 'water_supply'
        model_data = {
            'name': model.name,
            'interventions': [
                {"name": "water_asset_a", "location": "oxford"},
                {"name": "water_asset_b", "location": "oxford"},
                {"name": "water_asset_c", "location": "oxford"}
            ],
            'initial_conditions': [],
            'inputs': [],
            'outputs': [],
        }

        builder.add_model(model)
        builder.add_model_data(model, model_data)
        assert isinstance(builder.sos_model.models['water_supply'],
                          SectorModel)

        sos_model = builder.finish()
        assert isinstance(sos_model, SosModel)

        assert sos_model.timesteps == [2010, 2011, 2012]
        assert sos_model.sector_models == ['water_supply']
        assert sos_model.intervention_names == [
            "water_asset_a",
            "water_asset_b",
            "water_asset_c"
        ]

    def test_construct(self, get_config_data):
        """Test constructing from single dict config
        """
        config = get_config_data
        builder = SosModelBuilder()
        builder.construct(config)
        sos_model = builder.finish()

        assert isinstance(sos_model, SosModel)
        assert sos_model.sector_models == ['water_supply']
        assert isinstance(sos_model.models['water_supply'], SectorModel)
        assert sos_model.timesteps == [2010, 2011, 2012]

    def test_missing_planning_asset(self, get_config_data):
        config = get_config_data
        config["planning"] = [
            {
                "name": "test_intervention",
                "build_date": 2012
            }
        ]
        builder = SosModelBuilder()
        builder.construct(config)

        with raises(AssertionError) as ex:
            builder.finish()
        assert "Intervention 'test_intervention' in planning file not found" in str(ex.value)

    def test_missing_planning_timeperiod(self, get_config_data):
        config = get_config_data
        config["planning"] = [
            {
                "name": "test_intervention",
                "location": "UK",
                "build_date": 2025
            }
        ]
        config["sector_model_data"][0]["interventions"] = [
            {
                "name": "test_intervention",
                "location": "UK"
            }
        ]
        builder = SosModelBuilder()
        builder.construct(config)

        with raises(AssertionError) as ex:
            builder.finish()
        assert "Timeperiod '2025' in planning file not found" in str(ex.value)

    def test_scenario_dependency(self, get_config_data, setup_region_data):
        """Expect successful build with dependency on scenario data

        Should raise error if no spatial or temporal sets are defined
        """
        config = get_config_data
        config["sector_model_data"][0]["inputs"] = [
            {
                'name': 'raininess',
                'spatial_resolution': 'blobby',
                'temporal_resolution': 'mega'
            }
        ]

        builder = SosModelBuilder()
        builder.construct(config)

        with raises(ValueError):
            builder.finish()

        builder.load_region_sets({'blobby': setup_region_data['features']})

        interval_data = [
            {
                'id': 'ultra',
                'start': 'P0Y',
                'end': 'P1Y'
            }
        ]
        builder.load_interval_sets({'mega': interval_data})
        builder.finish()

    def test_build_valid_dependencies(self, one_dependency,
                                      get_config_data, setup_region_data):
        builder = SosModelBuilder()
        builder.construct(get_config_data)

        ws = WaterSupplySectorModel()
        ws.name = "water_supply_broken"
        ws.inputs = one_dependency
        builder.add_model(ws)

        with raises(AssertionError) as error:
            builder.finish()

        msg = "Missing dependency: water_supply_broken depends on macguffins produced" + \
              ", which is not supplied."
        assert str(error.value) == msg

    def test_cyclic_dependencies(self, setup_region_data):
        a_inputs = [
            {
                'name': 'b value',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual'
            }
        ]

        a_outputs = [
            {
                'name': 'a value',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual'
            }
        ]

        b_inputs = [
            {
                'name': 'a value',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual'
            }
        ]

        b_outputs = [
            {
                'name': 'b value',
                'spatial_resolution': 'LSOA',
                'temporal_resolution': 'annual'
            }
        ]

        builder = SosModelBuilder()
        builder.add_timesteps([2010])
        builder.add_planning([])
        builder.load_region_sets({'LSOA': setup_region_data['features']})
        interval_data = [{'id': 1, 'start': 'P0Y', 'end': 'P1Y'}]
        builder.load_interval_sets({'annual': interval_data})

        a_model = WaterSupplySectorModel()
        a_model.name = "a_model"
        a_model.inputs = a_inputs
        a_model.outputs = a_outputs
        builder.add_model(a_model)

        b_model = WaterSupplySectorModel()
        b_model.name = "b_model"
        b_model.inputs = b_inputs
        b_model.outputs = b_outputs
        builder.add_model(b_model)

        builder.finish()

    def test_nest_scenario_data(self, setup_country_data):
        data = {
            "mass": [
                {
                    'year': 2015,
                    'region': 'GB',
                    'interval': 'spring',
                    'value': 3,
                    'units': 'kg'
                },
                {
                    'year': 2015,
                    'region': 'GB',
                    'interval': 'summer',
                    'value': 5,
                    'units': 'kg'
                },
                {
                    'year': 2015,
                    'region': 'NI',
                    'interval': 'spring',
                    'value': 1,
                    'units': 'kg'
                },
                {
                    'year': 2016,
                    'region': 'GB',
                    'interval': 'spring',
                    'value': 4,
                    'units': 'kg'
                }
            ]
        }

        expected = {
            2015: {
                "mass": [
                    SpaceTimeValue('GB', 'spring', 3, 'kg'),
                    SpaceTimeValue('GB', 'summer', 5, 'kg'),
                    SpaceTimeValue('NI', 'spring', 1, 'kg')
                ]
            },
            2016: {
                "mass": [
                    SpaceTimeValue('GB', 'spring', 4, 'kg')
                ]
            }
        }

        builder = SosModelBuilder()
        interval_data = [
            {'id': 'winter', 'start': 'P0M', 'end': 'P2M'},
            {'id': 'spring', 'start': 'P2M', 'end': 'P5M'},
            {'id': 'summer', 'start': 'P5M', 'end': 'P8M'},
            {'id': 'autumn', 'start': 'P8M', 'end': 'P11M'},
            {'id': 'winter', 'start': 'P11M', 'end': 'P1Y'},
        ]
        builder.load_interval_sets({'annual': interval_data})
        builder.load_region_sets({'country': setup_country_data['features']})
        builder.add_resolution_mapping({
            'scenario': {
                'mass': {
                    'spatial_resolution': 'country',
                    'temporal_resolution': 'annual'
                }
            }
        })
        builder.add_scenario_data(data)

        actual = builder.sos_model._scenario_data

        assert actual[2015]["mass"] == [
            SpaceTimeValue("GB", "spring", 3, 'kg'),
            SpaceTimeValue('GB', 'summer', 5, 'kg'),
            SpaceTimeValue('NI', 'spring', 1, 'kg')
        ]
        assert actual == expected

        expected_value = [3, 5, 1]
        expected_region = ['GB', 'GB', 'NI']

        for exp_val, exp_reg, entry in zip(expected_value,
                                           expected_region,
                                           actual[2015]['mass']):
            assert entry.value == exp_val
            assert entry.region == exp_reg
            assert entry.units == 'kg'

    def test_scenario_data_defaults(self, setup_region_data):
        data = {
            "length": [
                {
                    'year': 2015,
                    'interval': 1,
                    'value': 3.14,
                    'units': 'm',
                    'region': 'oxford'
                }
            ]
        }

        expected = {2015: {"length": [SpaceTimeValue("oxford", 1, 3.14, 'm')]}}

        builder = SosModelBuilder()
        interval_data = [{'id': 1, 'start': 'P0Y', 'end': 'P1Y'}]
        builder.load_interval_sets({'annual': interval_data})
        builder.load_region_sets({'LSOA': setup_region_data['features']})
        builder.add_resolution_mapping({
            'scenario': {
                'length': {
                    'spatial_resolution': 'LSOA',
                    'temporal_resolution': 'annual'
                }
            }
        })
        builder.add_scenario_data(data)
        assert builder.sos_model._scenario_data == expected

    def test_scenario_data_missing_year(self, setup_region_data):
        data = {
            "length": [
                {
                    'value': 3.14,
                    'units': 'm'
                }
            ]
        }

        builder = SosModelBuilder()
        interval_data = [{'id': 1, 'start': 'P0Y', 'end': 'P1Y'}]
        builder.load_interval_sets({'annual': interval_data})
        builder.load_region_sets({'LSOA': setup_region_data['features']})
        builder.add_resolution_mapping({
            'scenario': {
                'length': {
                    'spatial_resolution': 'LSOA',
                    'temporal_resolution': 'annual'
                }
            }
        })

        msg = "Scenario data item missing year"
        with raises(ValueError) as ex:
            builder.add_scenario_data(data)
        assert msg in str(ex.value)

    def test_scenario_data_missing_param_mapping(self):
        data = {
            "length": [
                {
                    'value': 3.14,
                    'units': 'm',
                    'year': 2015
                }
            ]
        }

        builder = SosModelBuilder()

        msg = "Parameter length not registered in resolution mapping"
        with raises(ValueError) as ex:
            builder.add_scenario_data(data)
        assert msg in str(ex.value)

    def test_scenario_data_missing_param_region(self, setup_region_data):
        data = {
            "length": [
                {
                    'value': 3.14,
                    'units': 'm',
                    'region': 'missing',
                    'interval': 1,
                    'year': 2015
                }
            ]
        }

        builder = SosModelBuilder()
        interval_data = [{'id': 1, 'start': 'P0Y', 'end': 'P1Y'}]
        builder.load_interval_sets({'annual': interval_data})
        builder.load_region_sets({'LSOA': setup_region_data['features']})
        builder.add_resolution_mapping({
            'scenario': {
                'length': {
                    'spatial_resolution': 'LSOA',
                    'temporal_resolution': 'annual'
                }
            }
        })

        msg = "Region missing not defined in set LSOA for parameter length"
        with raises(ValueError) as ex:
            builder.add_scenario_data(data)
        assert msg in str(ex)

    def test_scenario_data_missing_param_interval(self, setup_region_data):
        data = {
            "length": [
                {
                    'value': 3.14,
                    'units': 'm',
                    'region': 'oxford',
                    'interval': 'missing',
                    'year': 2015
                }
            ]
        }

        builder = SosModelBuilder()
        interval_data = [{'id': 1, 'start': 'P0Y', 'end': 'P1Y'}]
        builder.load_interval_sets({'annual': interval_data})
        builder.load_region_sets({'LSOA': setup_region_data['features']})
        builder.add_resolution_mapping({
            'scenario': {
                'length': {
                    'spatial_resolution': 'LSOA',
                    'temporal_resolution': 'annual'
                }
            }
        })

        msg = "Interval missing not defined in set annual for parameter length"
        with raises(ValueError) as ex:
            builder.add_scenario_data(data)
        assert msg in str(ex)

    def test_inputs_property(self,
                             get_sos_model_only_scenario_dependencies):
        sos_model = get_sos_model_only_scenario_dependencies
        actual = sos_model.inputs

        expected = {'raininess': ['water_supply', 'water_supply_2']}

        assert isinstance(actual, dict)

        for key, value in expected.items():
            assert key in actual.keys()
            for entry in value:
                assert entry in actual[key]

    def test_outputs_property(self,
                              get_sos_model_only_scenario_dependencies):
        sos_model = get_sos_model_only_scenario_dependencies
        actual = sos_model.outputs

        expected = {'cost': ['water_supply']}

        assert isinstance(actual, dict)

        for key, value in expected.items():
            assert key in actual.keys()
            for entry in value:
                assert entry in actual[key]

    def test_single_dependency(self,
                               get_sos_model_with_model_dependency):
        sos_model = get_sos_model_with_model_dependency
        outputs = sos_model.outputs

        expected_outputs = {'water': ['water_supply']}

        assert isinstance(outputs, dict)

        for key, value in expected_outputs.items():
            assert key in outputs.keys()
            for entry in value:
                assert entry in outputs[key]

        inputs = sos_model.inputs

        expected_inputs = {
            'raininess': ['water_supply'],
            'water': ['water_supply_2']
        }

        assert isinstance(inputs, dict)

        for key, value in expected_inputs.items():
            assert key in inputs.keys()
            for entry in value:
                assert entry in inputs[key]

        sos_model.run()



"""Test custom pieces of YAML input/output
"""
from smif import SpaceTimeValue
from smif.data_layer.load import dump, load


def test_dump_space_time_value(setup_folder_structure):
    """Test custom YAML representation
    """
    test_file = str(setup_folder_structure.join("test.yaml"))
    data = SpaceTimeValue("Middlesborough", 0, 33, "GW")
    dump(data, test_file)
    with open(test_file, 'r') as fh:
        line = fh.readline()
        assert line.strip() == "!<SpaceTimeValue> [Middlesborough, 0, 33, GW]"


def test_load_space_time_value(setup_folder_structure):
    test_file = str(setup_folder_structure.join("test.yaml"))
    with open(test_file, 'w') as fh:
        fh.write("!<SpaceTimeValue> [Middlesborough, 0, 33, GW]")
    data = load(test_file)
    assert data == SpaceTimeValue("Middlesborough", 0, 33, "GW")



from pytest import raises
from smif import SpaceTimeValue, StateData


class TestSpaceTimeValue():
    """Test SpaceTimeValue object
    """
    def test_create(self):
        """Object creation and accessors
        """
        val = SpaceTimeValue("UK", "year", 180, "units")

        assert val.region == "UK"
        assert val.interval == "year"
        assert val.value == 180
        assert val.units == "units"
        expected = "SpaceTimeValue('UK', 'year', 180, 'units')"
        assert repr(val) == expected

    def test_equality(self):
        """Expect different objects with the same values to compare equal
        """
        val = SpaceTimeValue("UK", "year", 180, "units")
        same_val = SpaceTimeValue("UK", "year", 180, "units")

        assert val == same_val

    def test_addition(self):
        """Add values, all else being equal
        """
        val = SpaceTimeValue("UK", "year", 180, "units")
        other_val = SpaceTimeValue("UK", "year", 20, "units")

        new_val = val + other_val
        # add values
        assert new_val.value == 200
        # otherwise the same
        assert val.region == "UK"
        assert val.interval == "year"
        assert val.units == "units"

    def test_addition_invalid(self):
        """Fail to add values, if anything else is not eqal
        """
        val = SpaceTimeValue("UK", "year", 180, "units")
        other_val = SpaceTimeValue("France", "year", 20, "units")

        with raises(ValueError) as excinfo:
            val + other_val
        assert "Cannot add SpaceTimeValue of differing region" in str(excinfo)

        val = SpaceTimeValue("UK", "year", 180, "units")
        other_val = SpaceTimeValue("UK", "month", 20, "units")

        with raises(ValueError) as excinfo:
            val + other_val
        assert "Cannot add SpaceTimeValue of differing interval" in str(excinfo)

        val = SpaceTimeValue("UK", "year", 180, "units")
        other_val = SpaceTimeValue("UK", "year", 20, "other units")

        with raises(ValueError) as excinfo:
            val + other_val
        assert "Cannot add SpaceTimeValue of differing units" in str(excinfo)


class TestStateData():
    """Test StateData object
    """
    def test_create(self):
        """StateData creation and accessors
        """
        val = StateData(1, {'test': "data"})

        assert val.target == 1
        assert val.data == {'test': "data"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Implements example simulation model which can be run from the command line

Arguments
=========
raininess : int
    Sets the amount of rain

"""

from argparse import ArgumentParser
from water_supply import ExampleWaterSupplySimulationModel


def argparse():
    parser = ArgumentParser()
    parser.add_argument("--raininess",
                        type=int,
                        help="Sets the amount of rain")
    return parser.parse_args()


def main():
    args = argparse()
    water_supply = ExampleWaterSupplySimulationModel(args.raininess)
    results = water_supply.simulate()
    for key, val in results.items():
        print("{},{}".format(key, val))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Parse yaml config files, to construct sector models
"""
import yaml
from smif import SpaceTimeValue

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


def load(file_path):
    """Parse yaml config file into plain data (lists, dicts and simple values)

    Parameters
    ----------
    file_path : str
        The path of the configuration file to parse
    """
    with open(file_path, 'r') as file_handle:
        return yaml.load(file_handle, Loader=Loader)


def dump(data, file_path):
    """Write plain data to a file as yaml

    Parameters
    ----------
    file_path : str
        The path of the configuration file to write
    data
        Data to write (should be lists, dicts and simple values)
    """
    with open(file_path, 'w') as file_handle:
        return yaml.dump(data, file_handle, Dumper=Dumper)


def space_time_value_representer(dumper, data):
    """Dump custom yaml representation of SpaceTimeValue
    """
    return dumper.represent_sequence(
        "SpaceTimeValue", [
            data.region,
            data.interval,
            data.value,
            data.units
        ]
    )


yaml.add_representer(SpaceTimeValue, space_time_value_representer, Dumper=Dumper)


def space_time_value_constructor(loader, node):
    """Load ustom yaml representation of SpaceTimeValue
    """
    value = loader.construct_sequence(node)
    return SpaceTimeValue(
        value[0],
        value[1],
        value[2],
        value[3]
    )


yaml.add_constructor("SpaceTimeValue", space_time_value_constructor, Loader=Loader)



"""Energy demand dummy model
"""
from smif import SpaceTimeValue
from smif.sector_model import SectorModel


class EDMWrapper(SectorModel):
    """Energy model
    """
    def initialise(self, initial_conditions):
        pass

    def simulate(self, decisions, state, data):
        results = {
            "cost": [
                SpaceTimeValue('England', 1, 3, "million £"),
                SpaceTimeValue('Scotland', 1, 3, "million £"),
                SpaceTimeValue('Wales', 1, 3, "million £")
            ],
            "water_demand": [
                SpaceTimeValue('England', 1, 3, "Ml"),
                SpaceTimeValue('Scotland', 1, 3, "Ml"),
                SpaceTimeValue('Wales', 1, 3, "Ml")
            ],
        }
        return [], results

    def extract_obj(self, results):
        return 0



"""The decision module handles the three planning levels

Currently, only pre-specified planning is implemented.

"""

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class Planning:
    """
    Holds the list of planned interventions, where a single planned intervention
    is an intervention with a build date after which it will be included in the
    modelled systems.

    For example, a small pumping station might be built in
    Oxford in 2045::

            {
                'name': 'small_pumping_station',
                'build_date': 2045
            }



    Attributes
    ----------
    planned_interventions : list
        A list of pre-specified planned interventions
    """

    def __init__(self, planned_interventions=None):
        if planned_interventions is not None:
            self.planned_interventions = planned_interventions
        else:
            self.planned_interventions = []

    @property
    def names(self):
        """Returns the set of assets defined in the planned interventions
        """
        return {plan['name'] for plan in self.planned_interventions}

    @property
    def timeperiods(self):
        """Returns the set of build dates defined in the planned interventions
        """
        return {plan['build_date'] for plan in self.planned_interventions}
